# Copyright 2014 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging

from vmware_nsx.plugins.nsx_v.vshield import nsxv_edge_cfg_obj


LOG = logging.getLogger(__name__)


class NsxvLoadbalancer(nsxv_edge_cfg_obj.NsxvEdgeCfgObj):

    SERVICE_NAME = 'loadbalancer'

    def __init__(
            self,
            enabled=True,
            enable_service_insertion=False,
            acceleration_enabled=False):
        super(NsxvLoadbalancer, self).__init__()
        self.payload = {
            'enabled': enabled,
            'enableServiceInsertion': enable_service_insertion,
            'accelerationEnabled': acceleration_enabled}
        self.virtual_servers = {}

    def get_service_name(self):
        return self.SERVICE_NAME

    def add_virtual_server(self, virtual_server):
        self.virtual_servers[virtual_server.payload['name']] = virtual_server

    def del_virtual_server(self, name):
        self.virtual_servers.pop(name, None)

    def serializable_payload(self):
        virt_servers = []
        app_profiles = []
        app_rules = []
        pools = []
        monitors = []

        virt_id = 1
        app_prof_id = 1
        app_rule_id = 1
        pool_id = 1
        monitor_id = 1
        member_id = 1

        for virtual_server in self.virtual_servers.values():
            s_virt = virtual_server.payload.copy()
            s_virt['virtualServerId'] = 'virtualServer-%d' % virt_id
            virt_id += 1

            # Setup app profile
            s_app_prof = virtual_server.app_profile.payload.copy()
            s_app_prof['applicationProfileId'] = ('applicationProfile-%d' %
                                                  app_prof_id)
            app_profiles.append(s_app_prof)
            app_prof_id += 1

            # Bind virtual server to app profile
            s_virt['applicationProfileId'] = s_app_prof['applicationProfileId']

            # Setup app rules
            if virtual_server.app_rules.values():
                s_virt['applicationRuleId'] = []
                for app_rule in virtual_server.app_rules.values():
                    s_app_rule = app_rule.payload.copy()
                    s_app_rule['applicationRuleId'] = ('applicationRule-%d' %
                                                       app_rule_id)
                    app_rule_id += 1

                    # Add to LB object, bind to virtual server
                    app_rules.append(s_app_rule)
                    s_virt['applicationRuleId'].append(
                        s_app_rule['applicationRuleId'])

            # Setup pools
            s_pool = virtual_server.default_pool.payload.copy()
            s_pool['poolId'] = 'pool-%d' % pool_id
            pool_id += 1
            pools.append(s_pool)

            # Add pool members
            s_pool['member'] = []
            for member in virtual_server.default_pool.members.values():
                s_m = member.payload.copy()
                s_m['memberId'] = 'member-%d' % member_id
                member_id += 1
                s_pool['member'].append(s_m)

            # Bind pool to virtual server
            s_virt['defaultPoolId'] = s_pool['poolId']

            s_pool['monitorId'] = []
            # Add monitors
            for monitor in virtual_server.default_pool.monitors.values():
                s_mon = monitor.payload.copy()
                s_mon['monitorId'] = 'monitor-%d' % monitor_id
                monitor_id += 1

                s_pool['monitorId'].append(s_mon['monitorId'])

                monitors.append(s_mon)

            virt_servers.append(s_virt)

        payload = self.payload.copy()
        payload['applicationProfile'] = app_profiles
        if app_rules:
            payload['applicationRule'] = app_rules
        payload['monitor'] = monitors
        payload['pool'] = pools
        payload['virtualServer'] = virt_servers
        payload['featureType'] = 'loadbalancer_4.0'

        return payload

    @staticmethod
    def get_loadbalancer(vcns_obj, edge_id):
        edge_lb = nsxv_edge_cfg_obj.NsxvEdgeCfgObj.get_object(
            vcns_obj,
            edge_id,
            '%s/config' % NsxvLoadbalancer.SERVICE_NAME)

        lb_obj = NsxvLoadbalancer(
            edge_lb['enabled'],
            edge_lb['enableServiceInsertion'],
            edge_lb['accelerationEnabled'])

        # Construct loadbalancer objects
        for virt_srvr in edge_lb['virtualServer']:
            v_s = NsxvLBVirtualServer(
                virt_srvr['name'],
                virt_srvr['ipAddress'],
                virt_srvr['port'],
                virt_srvr['protocol'],
                virt_srvr['enabled'],
                virt_srvr['accelerationEnabled'],
                virt_srvr['connectionLimit'])

            # Find application profile objects, attach to virtual server
            for app_prof in edge_lb['applicationProfile']:
                if (virt_srvr['applicationProfileId']
                        == app_prof['applicationProfileId']):
                    a_p = NsxvLBAppProfile(
                        app_prof['name'],
                        app_prof['serverSslEnabled'],
                        app_prof['sslPassthrough'],
                        app_prof['template'],
                        app_prof['insertXForwardedFor'])

                    if app_prof.get('persistence'):
                        a_p.set_persistence(
                            True,
                            app_prof['persistence']['method'],
                            app_prof['persistence'].get('cookieName'),
                            app_prof['persistence'].get('cookieMode'),
                            app_prof['persistence'].get('expire'))

                    v_s.set_app_profile(a_p)

            # Find default pool, attach to virtual server
            for pool in edge_lb['pool']:
                if virt_srvr['defaultPoolId'] == pool['poolId']:
                    p = NsxvLBPool(
                        pool['name'],
                        pool['algorithm'],
                        pool['transparent'])

                    # Add pool members to pool
                    for member in pool['member']:
                        m = NsxvLBPoolMember(
                            member['name'],
                            member['ipAddress'],
                            member['port'],
                            member['monitorPort'],
                            member['condition'],
                            member['weight'],
                            member['minConn'],
                            member['maxConn'])

                        p.add_member(m)

                    # Add monitors to pool
                    for mon in edge_lb['monitor']:
                        if mon['monitorId'] in pool['monitorId']:
                            m = NsxvLBMonitor(
                                mon['name'],
                                mon['interval'],
                                mon['maxRetries'],
                                mon['method'],
                                mon['timeout'],
                                mon['type'],
                                mon['url'])

                            p.add_monitor(m)

                    v_s.set_default_pool(p)

            # Add application rules to virtual server
            for rule in edge_lb['applicationRule']:
                if rule['applicationRuleId'] in virt_srvr['applicationRuleId']:
                    r = NsxvLBAppRule(
                        rule['name'],
                        rule['script'])

                    v_s.add_app_rule(r)

            lb_obj.add_virtual_server(v_s)

        return lb_obj


class NsxvLBAppProfile(object):
    def __init__(
            self,
            name,
            server_ssl_enabled=False,
            ssl_pass_through=False,
            template='TCP',
            insert_xff=False,
            client_ssl_cert=None,
            persist=False,
            persist_method='cookie',
            persist_cookie_name='JSESSIONID',
            persist_cookie_mode='insert',
            persist_expire=30):
        self.payload = {
            'name': name,
            'serverSslEnabled': server_ssl_enabled,
            'sslPassthrough': ssl_pass_through,
            'template': template,
            'insertXForwardedFor': insert_xff}

        if persist:
            self.payload['persistence'] = {
                'method': persist_method,
                'expire': persist_expire
            }
            if persist_cookie_mode == 'cookie':
                self.payload['persistence']['cookieMode'] = persist_cookie_mode
                self.payload['persistence']['cookieName'] = persist_cookie_name

        if client_ssl_cert:
            self.payload['clientSsl'] = {
                'clientAuth': 'ignore',
                'serviceCertificate': [client_ssl_cert]
            }

    def set_persistence(
            self,
            persist=False,
            persist_method='cookie',
            persist_cookie_name='JSESSIONID',
            persist_cookie_mode='insert',
            persist_expire=30):

        if persist:
            self.payload['persistence'] = {
                'method': persist_method,
                'expire': persist_expire
            }
            if persist_cookie_mode == 'cookie':
                self.payload['persistence']['cookieMode'] = persist_cookie_mode
                self.payload['persistence']['cookieName'] = persist_cookie_name

        else:
            self.payload.pop('persistence', None)


class NsxvLBAppRule(object):
    def __init__(self, name, script):
        self.payload = {
            'name': name,
            'script': script}


class NsxvLBVirtualServer(object):
    def __init__(
            self,
            name,
            ip_address,
            port=80,
            protocol='HTTP',
            enabled=True,
            acceleration_enabled=False,
            connection_limit=0,
            enable_service_insertion=False):
        self.payload = {
            'name': name,
            'ipAddress': ip_address,
            'port': port,
            'protocol': protocol,
            'enabled': enabled,
            'accelerationEnabled': acceleration_enabled,
            'connectionLimit': connection_limit,
            'enableServiceInsertion': enable_service_insertion}

        self.app_rules = {}
        self.app_profile = None
        self.default_pool = None

    def add_app_rule(self, app_rule):
        self.app_rules[app_rule.payload['name']] = app_rule

    def del_app_rule(self, name):
        self.app_rules.pop(name, None)

    def set_default_pool(self, pool):
        self.default_pool = pool

    def set_app_profile(self, app_profile):
        self.app_profile = app_profile


class NsxvLBMonitor(object):
    def __init__(
            self,
            name,
            interval=10,
            max_retries=3,
            method='GET',
            timeout=15,
            mon_type='http',
            url='/'):
        self.payload = {
            'name': name,
            'interval': interval,
            'maxRetries': max_retries,
            'method': method,
            'timeout': timeout,
            'type': mon_type,
            'url': url}


class NsxvLBPoolMember(object):
    def __init__(
            self,
            name,
            ip_address,
            port,
            monitor_port=None,
            condition='enabled',
            weight=1,
            min_conn=0,
            max_conn=0):

        self.payload = {
            'name': name,
            'ipAddress': ip_address,
            'port': port,
            'monitorPort': monitor_port,
            'condition': condition,
            'weight': weight,
            'minConn': min_conn,
            'maxConn': max_conn}


class NsxvLBPool(object):
    def __init__(
            self,
            name,
            algorithm='round-robin',
            transparent=False):
        self.payload = {
            'name': name,
            'algorithm': algorithm,
            'transparent': transparent}

        self.members = {}
        self.monitors = {}

    def add_member(self, member):
        self.members[member.payload['name']] = member

    def del_member(self, name):
        self.members.pop(name, None)

    def add_monitor(self, monitor):
        self.monitors[monitor.payload['name']] = monitor

    def del_monitor(self, name):
        self.monitors.pop(name, None)



# Copyright 2014 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Note: this import should be here in order to appear before NeutronDbPluginV2
#  in each of the plugins. If not: security-group/-rule will not have all the
# relevant extend dict registries.
from neutron.db.models import securitygroup  # noqa

from vmware_nsx.plugins.dvs import plugin as dvs
from vmware_nsx.plugins.nsx import plugin as nsx
from vmware_nsx.plugins.nsx_mh import plugin as nsx_mh
from vmware_nsx.plugins.nsx_v import plugin as nsx_v
from vmware_nsx.plugins.nsx_v3 import plugin as nsx_v3

NsxDvsPlugin = dvs.NsxDvsV2
NsxPlugin = nsx_mh.NsxPluginV2
NsxVPlugin = nsx_v.NsxVPluginV2
NsxV3Plugin = nsx_v3.NsxV3Plugin
NsxTVDPlugin = nsx.NsxTVDPlugin



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.tests import base

from vmware_nsx.plugins.nsx_v.vshield import vcns_driver
from vmware_nsx.services.lbaas.nsx_v import lbaas_common as lb_common


EDGE_ID = 'edge-x'
POOL_ID = 'b3dfb476-6fdf-4ddd-b6bd-e86ae78dc30b'


def firewall_section_maker(if_ip_list, vip_ip_list):
    return (
        '<section id="1132" name="LBaaS FW Rules"><rule><name>' + POOL_ID +
        '</name><action>allow</action><sources excluded="false"><source>'
        '<type>Ipv4Address</type><value>' + ','.join(if_ip_list) +
        '</value></source></sources><destinations excluded="false">'
        '<destination><type>Ipv4Address</type><value>' +
        ','.join(vip_ip_list) + '</value></destination></destinations></rule>'
        '</section>')


def if_maker(ip_list):
    intf = {
        'index': 1, 'name': 'internal1', 'addressGroups': {
            'addressGroups': [
                {'subnetPrefixLength': '24',
                 'secondaryAddresses': {
                     'ipAddress': ip_list,
                     'type': 'secondary_addresses'},
                 'primaryAddress': '10.0.0.1',
                 'subnetMask': '255.255.255.0'}]},
            'portgroupName': 'pg1234', 'label': 'vNic_1',
            'type': 'internal', 'portgroupId': 'virtualwire-31'}
    return intf


def if_list_maker(ip_list):
    if_list = {
        'vnics': [
            {'index': 0, 'name': 'external', 'addressGroups': {
                'addressGroups': [
                    {'subnetMask': '255.255.255.0',
                     'primaryAddress': '172.24.4.2',
                     'subnetPrefixLength': '24'}]},
             'portgroupName': 'VM Network', 'label': 'vNic_0',
             'type': 'uplink', 'portgroupId': 'network-13'},
            {'index': 1, 'name': 'internal1', 'addressGroups': {
                'addressGroups': [
                    {'subnetPrefixLength': '24',
                     'secondaryAddresses': {
                         'ipAddress': ip_list,
                         'type': 'secondary_addresses'},
                     'primaryAddress': '10.0.0.1',
                     'subnetMask': '255.255.255.0'}]},
             'portgroupName': 'pg1234',
             'label': 'vNic_1', 'type': 'internal',
             'portgroupId': 'virtualwire-31'},
            {'index': 2, 'name': 'vnic2',
             'addressGroups': {'addressGroups': []},
             'label': 'vNic_2', 'type': 'internal'},
            {'index': 3, 'name': 'vnic3',
             'addressGroups': {'addressGroups': []},
             'label': 'vNic_3', 'type': 'internal'}]}
    return if_list


class TestLbaasCommon(base.BaseTestCase):
    def setUp(self):
        super(TestLbaasCommon, self).setUp()
        callbacks = mock.Mock()
        callbacks.plugin = mock.Mock()
        self.edge_driver = vcns_driver.VcnsDriver(callbacks)
        self.edge_driver._lb_driver_prop = mock.Mock()

    def _mock_edge_driver_vcns(self, attr):
        return mock.patch.object(self.edge_driver.vcns, attr)

    def test_add_vip_as_secondary_ip(self):
        update_if = if_maker(['10.0.0.6', '10.0.0.8'])

        with self._mock_edge_driver_vcns('get_interfaces') as mock_get_if,\
                self._mock_edge_driver_vcns(
                    'update_interface') as mock_update_if:

            mock_get_if.return_value = (None, if_list_maker(['10.0.0.6']))

            lb_common.add_vip_as_secondary_ip(
                self.edge_driver.vcns, EDGE_ID, '10.0.0.8')
            mock_update_if.assert_called_with(EDGE_ID, update_if)

    def test_del_vip_as_secondary_ip(self):
        update_if = if_maker(['10.0.0.6'])

        with self._mock_edge_driver_vcns('get_interfaces') as mock_get_if,\
                self._mock_edge_driver_vcns(
                    'update_interface') as mock_update_if:

            mock_get_if.return_value = (None, if_list_maker(['10.0.0.6',
                                                             '10.0.0.8']))

            lb_common.del_vip_as_secondary_ip(
                self.edge_driver.vcns, EDGE_ID, '10.0.0.8')
            mock_update_if.assert_called_with(EDGE_ID, update_if)

    def test_get_edge_ip_addresses(self):
        get_if_list = if_list_maker(['10.0.0.6'])

        with mock.patch.object(self.edge_driver.vcns, 'get_interfaces',
                               return_value=(None, get_if_list)):
            ip_list = lb_common.get_edge_ip_addresses(self.edge_driver.vcns,
                                                      EDGE_ID)
            self.assertEqual(['172.24.4.2', '10.0.0.1'], ip_list)



# Copyright 2013 VMware, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import neutron.db.api as db

import decorator
from neutron_lib.api.definitions import portbindings as pbin
from neutron_lib import constants as lib_const
from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import excutils
import six
from sqlalchemy import func
from sqlalchemy.orm import exc
from sqlalchemy.sql import expression as expr

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.common import nsxv_constants
from vmware_nsx.db import db as nsx_db
from vmware_nsx.db import nsxv_models
from vmware_nsx.extensions import dhcp_mtu as ext_dhcp_mtu
from vmware_nsx.extensions import dns_search_domain as ext_dns_search_domain
from vmware_nsx.plugins.nsx_v import availability_zones as nsx_az
from vmware_nsx.plugins.nsx_v.vshield.common import constants

NsxvEdgeDhcpStaticBinding = nsxv_models.NsxvEdgeDhcpStaticBinding
LOG = logging.getLogger(__name__)


def add_nsxv_router_binding(session, router_id, vse_id, lswitch_id, status,
                            appliance_size=nsxv_constants.LARGE,
                            edge_type=nsxv_constants.SERVICE_EDGE,
                            availability_zone=None):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvRouterBinding(
            router_id=router_id,
            edge_id=vse_id,
            lswitch_id=lswitch_id,
            status=status,
            appliance_size=appliance_size,
            edge_type=edge_type,
            availability_zone=availability_zone)
        session.add(binding)
    return binding


@decorator.decorator
def warn_on_binding_status_error(f, *args, **kwargs):
    result = f(*args, **kwargs)
    if result is None:
        return
    # we support functions that return a single entry or a list
    if isinstance(result, list):
        bindings = result
    else:
        bindings = [result]

    for binding in bindings:
        if binding and binding['status'] == lib_const.ERROR:
            LOG.warning("Found NSXV router binding entry with status "
                        "%(status)s: router %(router)s, "
                        "edge %(edge)s, lswitch %(lswitch)s, "
                        "status description: %(desc)s ",
                        {'status': binding['status'],
                         'router': binding['router_id'],
                         'edge': binding['edge_id'],
                         'lswitch': binding['lswitch_id'],
                         'desc': binding['status_description']})
    return result


@warn_on_binding_status_error
def get_nsxv_router_binding(session, router_id):
    return session.query(nsxv_models.NsxvRouterBinding).filter_by(
        router_id=router_id).first()


@warn_on_binding_status_error
def get_nsxv_router_binding_by_edge(session, edge_id):
    return session.query(nsxv_models.NsxvRouterBinding).filter_by(
        edge_id=edge_id).first()


@warn_on_binding_status_error
def get_nsxv_router_bindings_by_edge(session, edge_id):
    return session.query(nsxv_models.NsxvRouterBinding).filter_by(
        edge_id=edge_id).all()


@warn_on_binding_status_error
def get_nsxv_router_bindings(session, filters=None,
                             like_filters=None):
    session = db.get_reader_session()
    query = session.query(nsxv_models.NsxvRouterBinding)
    return nsx_db._apply_filters_to_query(query, nsxv_models.NsxvRouterBinding,
                                          filters, like_filters).all()


def update_nsxv_router_binding(session, router_id, **kwargs):
    with session.begin(subtransactions=True):
        binding = (session.query(nsxv_models.NsxvRouterBinding).
                   filter_by(router_id=router_id).one())
        for key, value in six.iteritems(kwargs):
            binding[key] = value
    return binding


def delete_nsxv_router_binding(session, router_id):
    with session.begin(subtransactions=True):
        binding = (session.query(nsxv_models.NsxvRouterBinding).
                   filter_by(router_id=router_id).first())
        if binding:
            session.delete(binding)


def get_edge_availability_zone(session, edge_id):
    binding = get_nsxv_router_binding_by_edge(session, edge_id)
    if binding:
        return binding['availability_zone']


def get_router_availability_zone(session, router_id):
    binding = get_nsxv_router_binding(session, router_id)
    if binding:
        return binding['availability_zone']


def clean_edge_router_binding(session, edge_id):
    with session.begin(subtransactions=True):
        (session.query(nsxv_models.NsxvRouterBinding).
         filter_by(edge_id=edge_id).delete())


def get_edge_vnic_bindings_with_networks(session):
    query = session.query(nsxv_models.NsxvEdgeVnicBinding)
    return query.filter(
        nsxv_models.NsxvEdgeVnicBinding.network_id != expr.null()).all()


def get_edge_vnic_binding(session, edge_id, network_id):
    return session.query(nsxv_models.NsxvEdgeVnicBinding).filter_by(
        edge_id=edge_id, network_id=network_id).first()


def get_edge_vnic_bindings_by_edge(session, edge_id):
    query = session.query(nsxv_models.NsxvEdgeVnicBinding)
    return query.filter(
        nsxv_models.NsxvEdgeVnicBinding.edge_id == edge_id,
        nsxv_models.NsxvEdgeVnicBinding.network_id != expr.null()).all()


def get_edge_vnic_bindings_by_int_lswitch(session, lswitch_id):
    return session.query(nsxv_models.NsxvEdgeVnicBinding).filter_by(
        network_id=lswitch_id).all()


def create_edge_vnic_binding(session, edge_id, vnic_index,
                             network_id, tunnel_index=-1):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvEdgeVnicBinding(
            edge_id=edge_id,
            vnic_index=vnic_index,
            tunnel_index=tunnel_index,
            network_id=network_id)
        session.add(binding)
    return binding


def delete_edge_vnic_binding_by_network(session, edge_id, network_id):
    with session.begin(subtransactions=True):
        binding = (session.query(nsxv_models.NsxvEdgeVnicBinding).
                   filter_by(edge_id=edge_id, network_id=network_id).one())
        session.delete(binding)


def init_edge_vnic_binding(session, edge_id):
    """Init edge vnic binding to preallocated 10 available edge vnics."""

    with session.begin(subtransactions=True):
        for vnic_index in range(constants.MAX_VNIC_NUM)[1:]:
            start = (vnic_index - 1) * constants.MAX_TUNNEL_NUM
            stop = vnic_index * constants.MAX_TUNNEL_NUM
            for tunnel_index in range(start, stop):
                binding = nsxv_models.NsxvEdgeVnicBinding(
                    edge_id=edge_id,
                    vnic_index=vnic_index,
                    tunnel_index=tunnel_index + 1)
                session.add(binding)


def clean_edge_vnic_binding(session, edge_id):
    """Clean edge vnic binding."""

    with session.begin(subtransactions=True):
        (session.query(nsxv_models.NsxvEdgeVnicBinding).
         filter_by(edge_id=edge_id).delete())


def allocate_edge_vnic(session, edge_id, network_id):
    """Allocate an available edge vnic to network."""

    with session.begin(subtransactions=True):
        bindings = (session.query(nsxv_models.NsxvEdgeVnicBinding).
                    filter_by(edge_id=edge_id, network_id=None).all())
        for binding in bindings:
            if binding['tunnel_index'] % constants.MAX_TUNNEL_NUM == 1:
                binding['network_id'] = network_id
                session.add(binding)
                return binding
    msg = (_("Edge VNIC: Failed to allocate one available vnic on edge_id: "
             ":%(edge_id)s to network_id: %(network_id)s") %
           {'edge_id': edge_id, 'network_id': network_id})
    LOG.error(msg)
    raise nsx_exc.NsxPluginException(err_msg=msg)


def allocate_edge_vnic_with_tunnel_index(session, edge_id, network_id,
                                         availability_zone):
    """Allocate an available edge vnic with tunnel index to network."""

    # TODO(berlin): temporary solution to let metadata and dhcp use
    # different vnics
    int_net = get_nsxv_internal_network(
        session, constants.InternalEdgePurposes.INTER_EDGE_PURPOSE,
        availability_zone)
    metadata_net_id = int_net['network_id'] if int_net else None

    with session.begin(subtransactions=True):
        query = session.query(nsxv_models.NsxvEdgeVnicBinding)
        query = query.filter(
            nsxv_models.NsxvEdgeVnicBinding.edge_id == edge_id,
            nsxv_models.NsxvEdgeVnicBinding.network_id == expr.null())
        if metadata_net_id:
            vnic_binding = get_edge_vnic_binding(
                session, edge_id, metadata_net_id)
            if vnic_binding:
                vnic_index = vnic_binding.vnic_index
                query = query.filter(
                    nsxv_models.NsxvEdgeVnicBinding.vnic_index != vnic_index)

        binding = query.first()
        if not binding:
            msg = (_("Failed to allocate one available vnic on edge_id: "
                     ":%(edge_id)s to network_id: %(network_id)s") %
                   {'edge_id': edge_id, 'network_id': network_id})
            LOG.error(msg)
            raise nsx_exc.NsxPluginException(err_msg=msg)
        binding['network_id'] = network_id
        session.add(binding)
    return binding


def allocate_specific_edge_vnic(session, edge_id, vnic_index,
                                tunnel_index, network_id):
    """Allocate an specific edge vnic to network."""

    with session.begin(subtransactions=True):
        binding = (session.query(nsxv_models.NsxvEdgeVnicBinding).
                   filter_by(edge_id=edge_id,
                             vnic_index=vnic_index,
                             tunnel_index=tunnel_index).one())
        binding['network_id'] = network_id
        session.add(binding)
    return binding


def get_dhcp_edge_network_binding(session, network_id):
    with session.begin(subtransactions=True):
        dhcp_router_edges = [binding['edge_id']
                             for binding in get_nsxv_router_bindings(session)
                             if binding['router_id'].startswith(
                                 constants.DHCP_EDGE_PREFIX)]
        bindings = (session.query(nsxv_models.NsxvEdgeVnicBinding).
                    filter_by(network_id=network_id))
        for binding in bindings:
            edge_id = binding['edge_id']
            if edge_id in dhcp_router_edges:
                return binding


def free_edge_vnic_by_network(session, edge_id, network_id):
    """Free an edge vnic."""

    with session.begin(subtransactions=True):
        binding = (session.query(nsxv_models.NsxvEdgeVnicBinding).
                   filter_by(edge_id=edge_id, network_id=network_id).one())
        binding['network_id'] = None
        session.add(binding)
    return binding


def _create_edge_dhcp_static_binding(session, edge_id, mac_address,
                                     binding_id):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvEdgeDhcpStaticBinding(
            edge_id=edge_id,
            mac_address=mac_address,
            binding_id=binding_id)
        session.add(binding)
    return binding


def create_edge_dhcp_static_binding(session, edge_id, mac_address, binding_id):
    try:
        return _create_edge_dhcp_static_binding(session, edge_id, mac_address,
                                                binding_id)
    except db_exc.DBDuplicateEntry:
        LOG.warning('Conflicting DHCP binding entry for '
                    '%(edge_id)s:%(mac_address)s. Overwriting!',
                    {'edge_id': edge_id, 'mac_address': mac_address})
        delete_edge_dhcp_static_binding(session, edge_id, mac_address)
        return _create_edge_dhcp_static_binding(session, edge_id, mac_address,
                                                binding_id)


def get_edge_dhcp_static_binding(session, edge_id, mac_address):
    return session.query(nsxv_models.NsxvEdgeDhcpStaticBinding).filter_by(
        edge_id=edge_id, mac_address=mac_address).first()


def get_dhcp_static_bindings_by_edge(session, edge_id):
    return session.query(nsxv_models.NsxvEdgeDhcpStaticBinding).filter_by(
        edge_id=edge_id).all()


def delete_edge_dhcp_static_binding(session, edge_id, mac_address):
    with session.begin(subtransactions=True):
        session.query(nsxv_models.NsxvEdgeDhcpStaticBinding).filter_by(
            edge_id=edge_id, mac_address=mac_address).delete()


def delete_edge_dhcp_static_binding_id(session, edge_id, binding_id):
    with session.begin(subtransactions=True):
        session.query(nsxv_models.NsxvEdgeDhcpStaticBinding).filter_by(
            edge_id=edge_id, binding_id=binding_id).delete()


def get_nsxv_dhcp_bindings_count_per_edge(session):
    return (
        session.query(
            NsxvEdgeDhcpStaticBinding.edge_id,
            func.count(NsxvEdgeDhcpStaticBinding.mac_address)).group_by(
            NsxvEdgeDhcpStaticBinding.edge_id).all())


def clean_edge_dhcp_static_bindings_by_edge(session, edge_id):
    with session.begin(subtransactions=True):
        session.query(nsxv_models.NsxvEdgeDhcpStaticBinding).filter_by(
            edge_id=edge_id).delete()


def create_nsxv_internal_network(session, network_purpose,
                                 availability_zone, network_id):
    with session.begin(subtransactions=True):
        try:
            network = nsxv_models.NsxvInternalNetworks(
                network_purpose=network_purpose,
                network_id=network_id,
                availability_zone=availability_zone)
            session.add(network)
        except db_exc.DBDuplicateEntry:
            with excutils.save_and_reraise_exception():
                LOG.exception("Duplicate internal network for purpose "
                              "%(p)s and availabiltiy zone %(az)s",
                              {'p': network_purpose,
                              'az': availability_zone})


def get_nsxv_internal_network(session, network_purpose, availability_zone,
                              default_fallback=True):
    with session.begin(subtransactions=True):
        net_list = (session.query(nsxv_models.NsxvInternalNetworks).
                    filter_by(network_purpose=network_purpose,
                              availability_zone=availability_zone).all())
        if net_list:
            # Should have only one results as purpose+az are the keys
            return net_list[0]
        elif default_fallback and availability_zone != nsx_az.DEFAULT_NAME:
            # try the default availability zone, since this zone does not
            # have his own internal edge
            net_list = (session.query(nsxv_models.NsxvInternalNetworks).
                        filter_by(network_purpose=network_purpose,
                                  availability_zone=nsx_az.DEFAULT_NAME).all())
            if net_list:
                return net_list[0]


def get_nsxv_internal_network_for_az(session, network_purpose,
                                     availability_zone):
    return get_nsxv_internal_network(session, network_purpose,
                                     availability_zone,
                                     default_fallback=False)


def get_nsxv_internal_networks(session, network_purpose):
    with session.begin(subtransactions=True):
        return (session.query(nsxv_models.NsxvInternalNetworks).
                filter_by(network_purpose=network_purpose).all())


def get_nsxv_internal_network_by_id(session, network_id):
    with session.begin(subtransactions=True):
        return (session.query(nsxv_models.NsxvInternalNetworks).
                filter_by(network_id=network_id).first())


def delete_nsxv_internal_network(session, network_purpose, network_id):
    with session.begin(subtransactions=True):
        return (session.query(nsxv_models.NsxvInternalNetworks).
                filter_by(network_purpose=network_purpose,
                          network_id=network_id).delete())


def create_nsxv_internal_edge(session, ext_ip_address, purpose, router_id):
    with session.begin(subtransactions=True):
        try:
            internal_edge = nsxv_models.NsxvInternalEdges(
                ext_ip_address=ext_ip_address,
                purpose=purpose,
                router_id=router_id)
            session.add(internal_edge)
        except db_exc.DBDuplicateEntry:
            with excutils.save_and_reraise_exception():
                LOG.exception("Duplicate internal Edge IP %s",
                              ext_ip_address)


def get_nsxv_internal_edge(session, ext_ip_address):
    with session.begin(subtransactions=True):
        return (session.query(nsxv_models.NsxvInternalEdges).
                filter_by(ext_ip_address=ext_ip_address).all())


def update_nsxv_internal_edge(session, ext_ip_address, router_id):
    with session.begin(subtransactions=True):
        edges = get_nsxv_internal_edge(session, ext_ip_address)

        for edge in edges:
            edge['router_id'] = router_id


def get_nsxv_internal_edges_by_purpose(session, purpose):
    with session.begin(subtransactions=True):
        return (session.query(nsxv_models.NsxvInternalEdges).
                filter_by(purpose=purpose).all())


def get_nsxv_internal_edge_by_router(session, router_id):
    with session.begin(subtransactions=True):
        return (session.query(nsxv_models.NsxvInternalEdges).
                filter_by(router_id=router_id).first())


def delete_nsxv_internal_edge(session, ext_ip_address):
    with session.begin(subtransactions=True):
        return (session.query(nsxv_models.NsxvInternalEdges).
                filter_by(ext_ip_address=ext_ip_address).delete())


def add_neutron_nsx_section_mapping(session, neutron_id, section_id):
    with session.begin(subtransactions=True):
        mapping = nsxv_models.NsxvSecurityGroupSectionMapping(
            neutron_id=neutron_id, ip_section_id=section_id)
        session.add(mapping)
    return mapping


def add_neutron_nsx_rule_mapping(session, neutron_id, nsx_rule_id):
    with session.begin(subtransactions=True):
        mapping = nsxv_models.NsxvRuleMapping(neutron_id=neutron_id,
                                              nsx_rule_id=nsx_rule_id)
        session.add(mapping)
    return mapping


def add_neutron_nsx_port_vnic_mapping(session, neutron_id, nsx_id):
    with session.begin(subtransactions=True):
        mapping = nsxv_models.NsxvPortVnicMapping(
            neutron_id=neutron_id, nsx_id=nsx_id)
        session.add(mapping)
    return mapping


def get_nsx_section(session, neutron_id):
    try:
        mapping = (session.query(nsxv_models.NsxvSecurityGroupSectionMapping).
                   filter_by(neutron_id=neutron_id).
                   one())
        return mapping
    except exc.NoResultFound:
        LOG.debug("NSX identifiers for neutron security group %s not yet "
                  "stored in Neutron DB", neutron_id)


def delete_neutron_nsx_section_mapping(session, neutron_id):
    with session.begin(subtransactions=True):
        return (session.query(nsxv_models.NsxvSecurityGroupSectionMapping).
                filter_by(neutron_id=neutron_id).delete())


def get_nsx_rule_id(session, neutron_id):
    try:
        mapping = (session.query(nsxv_models.NsxvRuleMapping).
                   filter_by(neutron_id=neutron_id).
                   one())
        return mapping['nsx_rule_id']
    except exc.NoResultFound:
        LOG.debug("NSX identifiers for neutron rule %s not yet "
                  "stored in Neutron DB", neutron_id)


def get_nsx_vnic_id(session, neutron_id):
    try:
        mapping = (session.query(nsxv_models.NsxvPortVnicMapping).
                   filter_by(neutron_id=neutron_id).
                   one())
        return mapping['nsx_id']
    except exc.NoResultFound:
        LOG.debug("NSX identifiers for neutron port %s not yet "
                  "stored in Neutron DB", neutron_id)


def get_network_bindings(session, network_id):
    session = session or db.get_reader_session()
    return (session.query(nsxv_models.NsxvTzNetworkBinding).
            filter_by(network_id=network_id).
            all())


def get_network_bindings_by_vlanid_and_physical_net(session, vlan_id,
                                                    phy_uuid):
    session = session or db.get_reader_session()
    return (session.query(nsxv_models.NsxvTzNetworkBinding).
            filter_by(vlan_id=vlan_id, phy_uuid=phy_uuid).
            all())


def get_network_bindings_by_ids(session, vlan_id, phy_uuid):
    return get_network_bindings_by_vlanid_and_physical_net(
        session, vlan_id, phy_uuid)


def get_network_bindings_by_physical_net(session, phy_uuid):
    session = session or db.get_reader_session()
    return (session.query(nsxv_models.NsxvTzNetworkBinding).
            filter_by(phy_uuid=phy_uuid).
            all())


def get_network_bindings_by_physical_net_and_type(session, phy_uuid,
                                                  binding_type):
    session = session or db.get_reader_session()
    return (session.query(nsxv_models.NsxvTzNetworkBinding).
            filter_by(phy_uuid=phy_uuid,
                      binding_type=binding_type).
            all())


def delete_network_bindings(session, network_id):
    return (session.query(nsxv_models.NsxvTzNetworkBinding).
            filter_by(network_id=network_id).delete())


def add_network_binding(session, network_id, binding_type, phy_uuid, vlan_id):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvTzNetworkBinding(network_id, binding_type,
                                                   phy_uuid, vlan_id)
        session.add(binding)
    return binding


def get_network_bindings_by_vlanid(session, vlan_id):
    session = session or db.get_reader_session()
    return (session.query(nsxv_models.NsxvTzNetworkBinding).
            filter_by(vlan_id=vlan_id).
            all())


def update_network_binding_phy_uuid(session, network_id, binding_type,
                                    vlan_id, phy_uuid):
    with session.begin(subtransactions=True):
        bindings = (session.query(nsxv_models.NsxvTzNetworkBinding).filter_by(
            vlan_id=vlan_id,
            network_id=network_id,
            binding_type=binding_type).all())
        for binding in bindings:
            binding['phy_uuid'] = phy_uuid


#
# Edge Firewall binding methods
#
def add_nsxv_edge_firewallrule_binding(session, map_info):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvEdgeFirewallRuleBinding(
            rule_id=map_info['rule_id'],
            rule_vse_id=map_info['rule_vseid'],
            edge_id=map_info['edge_id'])
        session.add(binding)
    return binding


def delete_nsxv_edge_firewallrule_binding(session, id):
    with session.begin(subtransactions=True):
        if not (session.query(nsxv_models.NsxvEdgeFirewallRuleBinding).
                filter_by(rule_id=id).delete()):
            msg = _("Rule Resource binding with id:%s not found!") % id
            raise nsx_exc.NsxPluginException(err_msg=msg)


def get_nsxv_edge_firewallrule_binding(session, id, edge_id):
    with session.begin(subtransactions=True):
        return (session.query(nsxv_models.NsxvEdgeFirewallRuleBinding).
                filter_by(rule_id=id, edge_id=edge_id).first())


def get_nsxv_edge_firewallrule_binding_by_vseid(
        session, edge_id, rule_vseid):
    with session.begin(subtransactions=True):
        try:
            return (session.query(nsxv_models.NsxvEdgeFirewallRuleBinding).
                    filter_by(edge_id=edge_id, rule_vse_id=rule_vseid).one())
        except exc.NoResultFound:
            return


def cleanup_nsxv_edge_firewallrule_binding(session, edge_id):
    with session.begin(subtransactions=True):
        session.query(
            nsxv_models.NsxvEdgeFirewallRuleBinding).filter_by(
                edge_id=edge_id).delete()


def map_spoofguard_policy_for_network(session, network_id, policy_id):
    with session.begin(subtransactions=True):
        mapping = nsxv_models.NsxvSpoofGuardPolicyNetworkMapping(
            network_id=network_id, policy_id=policy_id)
        session.add(mapping)
    return mapping


def get_spoofguard_policy_id(session, network_id):
    try:
        mapping = (session.query(
            nsxv_models.NsxvSpoofGuardPolicyNetworkMapping).
            filter_by(network_id=network_id).one())
        return mapping['policy_id']
    except exc.NoResultFound:
        LOG.debug("SpoofGuard Policy for network %s was not found",
                  network_id)


def get_nsxv_spoofguard_policy_network_mappings(session, filters=None,
                                                like_filters=None):
    session = db.get_reader_session()
    query = session.query(nsxv_models.NsxvSpoofGuardPolicyNetworkMapping)
    return nsx_db._apply_filters_to_query(
               query, nsxv_models.NsxvSpoofGuardPolicyNetworkMapping,
               filters, like_filters).all()


def add_nsxv_lbaas_loadbalancer_binding(
        session, loadbalancer_id, edge_id, edge_fw_rule_id, vip_address):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvLbaasLoadbalancerBinding(
            loadbalancer_id=loadbalancer_id,
            edge_id=edge_id,
            edge_fw_rule_id=edge_fw_rule_id,
            vip_address=vip_address)
        session.add(binding)
    return binding


def get_nsxv_lbaas_loadbalancer_binding(session, loadbalancer_id):
    try:
        return session.query(
            nsxv_models.NsxvLbaasLoadbalancerBinding).filter_by(
            loadbalancer_id=loadbalancer_id).one()
    except exc.NoResultFound:
        return


def get_nsxv_lbaas_loadbalancer_binding_by_edge(session, edge_id):
    return session.query(
        nsxv_models.NsxvLbaasLoadbalancerBinding).filter_by(
        edge_id=edge_id).all()


def del_nsxv_lbaas_loadbalancer_binding(session, loadbalancer_id):
    return (session.query(nsxv_models.NsxvLbaasLoadbalancerBinding).
            filter_by(loadbalancer_id=loadbalancer_id).delete())


def add_nsxv_lbaas_listener_binding(session, loadbalancer_id, listener_id,
                                    app_profile_id, vse_id):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvLbaasListenerBinding(
            loadbalancer_id=loadbalancer_id,
            listener_id=listener_id,
            app_profile_id=app_profile_id,
            vse_id=vse_id)
        session.add(binding)
    return binding


def get_nsxv_lbaas_listener_binding(session, loadbalancer_id, listener_id):
    try:
        return session.query(
            nsxv_models.NsxvLbaasListenerBinding).filter_by(
            loadbalancer_id=loadbalancer_id, listener_id=listener_id).one()
    except exc.NoResultFound:
        return


def del_nsxv_lbaas_listener_binding(session, loadbalancer_id, listener_id):
    return (session.query(nsxv_models.NsxvLbaasListenerBinding).
            filter_by(loadbalancer_id=loadbalancer_id,
                      listener_id=listener_id).delete())


def add_nsxv_lbaas_pool_binding(session, loadbalancer_id, pool_id,
                                edge_pool_id):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvLbaasPoolBinding(
            loadbalancer_id=loadbalancer_id,
            pool_id=pool_id,
            edge_pool_id=edge_pool_id)
        session.add(binding)
    return binding


def get_nsxv_lbaas_pool_binding(session, loadbalancer_id, pool_id):
    try:
        return session.query(
            nsxv_models.NsxvLbaasPoolBinding).filter_by(
            loadbalancer_id=loadbalancer_id,
            pool_id=pool_id).one()
    except exc.NoResultFound:
        return


def del_nsxv_lbaas_pool_binding(session, loadbalancer_id, pool_id):
    return (session.query(nsxv_models.NsxvLbaasPoolBinding).
            filter_by(loadbalancer_id=loadbalancer_id,
                      pool_id=pool_id).delete())


def add_nsxv_lbaas_monitor_binding(session, loadbalancer_id, pool_id, hm_id,
                                   edge_id, edge_mon_id):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvLbaasMonitorBinding(
            loadbalancer_id=loadbalancer_id,
            pool_id=pool_id,
            hm_id=hm_id,
            edge_id=edge_id,
            edge_mon_id=edge_mon_id)
        session.add(binding)
    return binding


def get_nsxv_lbaas_monitor_binding(session, loadbalancer_id, pool_id, hm_id,
                                   edge_id):
    try:
        return session.query(
            nsxv_models.NsxvLbaasMonitorBinding).filter_by(
            loadbalancer_id=loadbalancer_id,
            pool_id=pool_id,
            hm_id=hm_id,
            edge_id=edge_id).one()
    except exc.NoResultFound:
        return


def del_nsxv_lbaas_monitor_binding(session, loadbalancer_id, pool_id, hm_id,
                                   edge_id):
    return (session.query(nsxv_models.NsxvLbaasMonitorBinding).
            filter_by(loadbalancer_id=loadbalancer_id,
                      pool_id=pool_id,
                      hm_id=hm_id,
                      edge_id=edge_id).delete())


def add_nsxv_lbaas_certificate_binding(session, cert_id, edge_id,
                                       edge_cert_id):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvLbaasCertificateBinding(
            cert_id=cert_id,
            edge_id=edge_id,
            edge_cert_id=edge_cert_id)
        session.add(binding)
    return binding


def get_nsxv_lbaas_certificate_binding(session, cert_id, edge_id):
    try:
        return session.query(
            nsxv_models.NsxvLbaasCertificateBinding).filter_by(
            cert_id=cert_id,
            edge_id=edge_id).one()
    except exc.NoResultFound:
        return


def del_nsxv_lbaas_certificate_binding(session, cert_id, edge_id):
    return (session.query(nsxv_models.NsxvLbaasCertificateBinding).
            filter_by(cert_id=cert_id,
                      edge_id=edge_id).delete())


def add_nsxv_lbaas_l7policy_binding(session, policy_id, edge_id,
                                    edge_app_rule_id):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvLbaasL7PolicyBinding(
            policy_id=policy_id,
            edge_id=edge_id,
            edge_app_rule_id=edge_app_rule_id)
        session.add(binding)
    return binding


def get_nsxv_lbaas_l7policy_binding(session, policy_id):
    try:
        return session.query(
            nsxv_models.NsxvLbaasL7PolicyBinding).filter_by(
            policy_id=policy_id).one()
    except exc.NoResultFound:
        return


def del_nsxv_lbaas_l7policy_binding(session, policy_id):
    try:
        return (session.query(nsxv_models.NsxvLbaasL7PolicyBinding).
                filter_by(policy_id=policy_id).delete())
    except exc.NoResultFound:
        return


def add_nsxv_subnet_ext_attributes(session, subnet_id,
                                   dns_search_domain=None,
                                   dhcp_mtu=None):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvSubnetExtAttributes(
            subnet_id=subnet_id,
            dns_search_domain=dns_search_domain,
            dhcp_mtu=dhcp_mtu)
        session.add(binding)
    return binding


def get_nsxv_subnet_ext_attributes(session, subnet_id):
    try:
        return session.query(
            nsxv_models.NsxvSubnetExtAttributes).filter_by(
            subnet_id=subnet_id).one()
    except exc.NoResultFound:
        return


def update_nsxv_subnet_ext_attributes(session, subnet_id,
                                      dns_search_domain=None,
                                      dhcp_mtu=None):
    with session.begin(subtransactions=True):
        binding = (session.query(nsxv_models.NsxvSubnetExtAttributes).
                   filter_by(subnet_id=subnet_id).one())
        binding[ext_dns_search_domain.DNS_SEARCH_DOMAIN] = dns_search_domain
        binding[ext_dhcp_mtu.DHCP_MTU] = dhcp_mtu
    return binding


def add_nsxv_port_ext_attributes(session, port_id,
                                 vnic_type=pbin.VNIC_NORMAL):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvPortExtAttributes(
            port_id=port_id,
            vnic_type=vnic_type)
        session.add(binding)
    return binding


def get_nsxv_ext_attr_port_vnic_type(session, port_id):
    try:
        binding = session.query(nsxv_models.NsxvPortExtAttributes).filter_by(
            port_id=port_id).one()
        return binding['vnic_type']
    except exc.NoResultFound:
        return pbin.VNIC_NORMAL


def update_nsxv_port_ext_attributes(session, port_id,
                                    vnic_type=pbin.VNIC_NORMAL):
    try:
        binding = session.query(
            nsxv_models.NsxvPortExtAttributes).filter_by(
            port_id=port_id).one()
        binding['vnic_type'] = vnic_type
        return binding
    except exc.NoResultFound:
        return add_nsxv_port_ext_attributes(
            session, port_id, vnic_type=vnic_type)


def add_nsxv_bgp_speaker_binding(session, edge_id, speaker_id,
                                 bgp_identifier):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvBgpSpeakerBinding(
            edge_id=edge_id,
            bgp_speaker_id=speaker_id,
            bgp_identifier=bgp_identifier)
        session.add(binding)
        return binding


def get_nsxv_bgp_speaker_binding(session, edge_id):
    try:
        binding = (session.query(nsxv_models.NsxvBgpSpeakerBinding).
                   filter_by(edge_id=edge_id).
                   one())
        return binding
    except exc.NoResultFound:
        LOG.debug("No dynamic routing enabled on edge %s.", edge_id)


def get_nsxv_bgp_speaker_bindings(session, speaker_id):
    try:
        return (session.query(nsxv_models.NsxvBgpSpeakerBinding).
                filter_by(bgp_speaker_id=speaker_id).all())
    except exc.NoResultFound:
        return []


def delete_nsxv_bgp_speaker_binding(session, edge_id):
    binding = session.query(
        nsxv_models.NsxvBgpSpeakerBinding).filter_by(edge_id=edge_id)
    if binding:
        binding.delete()


def add_nsxv_bgp_peer_edge_binding(session, peer_id, edge_id):
    with session.begin(subtransactions=True):
        binding = nsxv_models.NsxvBgpPeerEdgeBinding(edge_id=edge_id,
                                                     peer_id=peer_id)
        session.add(binding)
        return binding


def get_nsxv_bgp_peer_edge_binding(session, peer_id):
    try:
        binding = (session.query(nsxv_models.NsxvBgpPeerEdgeBinding).
                   filter_by(peer_id=peer_id).one())
        return binding
    except exc.NoResultFound:
        pass



# Copyright 2015 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron.db import l3_db
from neutron_lib.callbacks import registry
from neutron_lib import exceptions
from oslo_log import log as logging
from oslo_utils import uuidutils

from vmware_nsx.common import locking
from vmware_nsx.common import nsxv_constants
from vmware_nsx.db import nsxv_db
from vmware_nsx.db import nsxv_models
from vmware_nsx.plugins.nsx_v.vshield.common import constants as vcns_const
from vmware_nsx.plugins.nsx_v.vshield import edge_utils
from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import formatters
import vmware_nsx.shell.admin.plugins.common.utils as admin_utils
import vmware_nsx.shell.admin.plugins.nsxv.resources.utils as utils
import vmware_nsx.shell.resources as shell


LOG = logging.getLogger(__name__)
nsxv = utils.get_nsxv_client()
_uuid = uuidutils.generate_uuid


def get_nsxv_backup_edges():
    edges = utils.get_nsxv_backend_edges()
    backup_edges = []
    edgeapi = utils.NeutronDbClient()
    for edge in edges:
        if edge['name'].startswith("backup-"):
            # Make sure it is really a backup edge
            edge_vnic_binds = nsxv_db.get_edge_vnic_bindings_by_edge(
                edgeapi.context.session, edge['id'])
            if not edge_vnic_binds:
                extend_edge_info(edge)
                backup_edges.append(edge)
    return backup_edges


def extend_edge_info(edge):
    """Add information from the nsxv-db, if available"""
    edgeapi = utils.NeutronDbClient()
    rtr_binding = nsxv_db.get_nsxv_router_binding_by_edge(
            edgeapi.context.session, edge['id'])
    if rtr_binding:
        edge['availability_zone'] = rtr_binding['availability_zone']
        edge['db_status'] = rtr_binding['status']


@admin_utils.output_header
def nsx_list_backup_edges(resource, event, trigger, **kwargs):
    """List backup edges"""
    backup_edges = get_nsxv_backup_edges()
    LOG.info(formatters.output_formatter(
        constants.BACKUP_EDGES, backup_edges,
        ['id', 'name', 'size', 'type', 'availability_zone', 'db_status']))


def _delete_backup_from_neutron_db(edge_id, router_id):
    # Remove bindings from Neutron DB
    edgeapi = utils.NeutronDbClient()
    nsxv_db.delete_nsxv_router_binding(
        edgeapi.context.session, router_id)
    if edge_id:
        nsxv_db.clean_edge_vnic_binding(edgeapi.context.session, edge_id)


def _delete_edge_from_nsx_and_neutron(edge_id, router_id):
    try:
        with locking.LockManager.get_lock(edge_id):
            # Delete from NSXv backend
            nsxv.delete_edge(edge_id)
            # Remove bindings from Neutron DB
            _delete_backup_from_neutron_db(edge_id, router_id)
            return True
    except Exception as expt:
        LOG.error("%s", str(expt))
        return False


def _nsx_delete_backup_edge(edge_id, all_backup_edges):
    """Delete a specific backup edge"""
    try:
        edge_result = nsxv.get_edge(edge_id)
    except exceptions.NeutronException as x:
        LOG.error("%s", str(x))
    else:
        # edge_result[0] is response status code
        # edge_result[1] is response body
        edge = edge_result[1]
        backup_edges = [e['id'] for e in all_backup_edges]
        if (not edge['name'].startswith('backup-')
            or edge['id'] not in backup_edges):
            LOG.error(
                'Edge: %s is not a backup edge; aborting delete',
                edge_id)
        else:
            return _delete_edge_from_nsx_and_neutron(edge_id, edge['name'])


def nsx_clean_backup_edge(resource, event, trigger, **kwargs):
    """Delete backup edge"""
    errmsg = ("Need to specify edge-id property. Add --property "
              "edge-id=<edge-id>")
    if not kwargs.get('property'):
        LOG.error("%s", errmsg)
        return
    properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
    edge_id = properties.get('edge-id')
    if not edge_id:
        LOG.error("%s", errmsg)
        return
    #ask for the user confirmation
    confirm = admin_utils.query_yes_no(
        "Do you want to delete edge: %s" % edge_id, default="no")
    if not confirm:
        LOG.info("Backup edge deletion aborted by user")
        return
    # delete the backup edge
    _nsx_delete_backup_edge(edge_id, get_nsxv_backup_edges())


def nsx_clean_all_backup_edges(resource, event, trigger, **kwargs):
    """Delete all backup edges"""
    backup_edges = get_nsxv_backup_edges()

    #ask for the user confirmation
    confirm = admin_utils.query_yes_no(
        "Do you want to delete %s backup edges?" % len(backup_edges),
        default="no")
    if not confirm:
        LOG.info("Backup edges deletion aborted by user")
        return

    deleted_cnt = 0
    for edge in backup_edges:
        # delete the backup edge
        if _nsx_delete_backup_edge(edge['id'], backup_edges):
            deleted_cnt = deleted_cnt + 1

    LOG.info('Done Deleting %s backup edges', deleted_cnt)


@admin_utils.output_header
def neutron_clean_backup_edge(resource, event, trigger, **kwargs):
    """Delete a backup edge from the neutron, and backend by it's name

    The name of the backup edge is the router-id column in the BD table
    nsxv_router_bindings, and it is also printed by list-mismatches
    """
    errmsg = ("Need to specify router-id property. Add --property "
              "router-id=<router-id>")
    if not kwargs.get('property'):
        LOG.error("%s", errmsg)
        return
    properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
    router_id = properties.get('router-id')
    if not router_id:
        LOG.error("%s", errmsg)
        return

    # look for the router-binding entry
    edgeapi = utils.NeutronDbClient()
    rtr_binding = nsxv_db.get_nsxv_router_binding(
            edgeapi.context.session, router_id)
    if not rtr_binding:
        LOG.error('Backup %s was not found in DB', router_id)
        return

    edge_id = rtr_binding['edge_id']
    if edge_id:
        # delete from backend too
        _delete_edge_from_nsx_and_neutron(edge_id, router_id)
    else:
        # delete only from DB
        _delete_backup_from_neutron_db(None, router_id)


@admin_utils.output_header
def nsx_list_name_mismatches(resource, event, trigger, **kwargs):
    edges = utils.get_nsxv_backend_edges()
    plugin_nsx_mismatch = []
    backend_edge_ids = []
    edgeapi = utils.NeutronDbClient()
    # Look for edges with the wrong names:
    for edge in edges:
        backend_edge_ids.append(edge['id'])
        rtr_binding = nsxv_db.get_nsxv_router_binding_by_edge(
                edgeapi.context.session, edge['id'])

        if (rtr_binding and
            edge['name'].startswith('backup-')
            and rtr_binding['router_id'] != edge['name']):
            plugin_nsx_mismatch.append(
                    {'edge_id': edge['id'],
                     'edge_name': edge['name'],
                     'router_id': rtr_binding['router_id']})

    LOG.info(formatters.output_formatter(
            constants.BACKUP_EDGES + ' with name mismatch:',
            plugin_nsx_mismatch,
            ['edge_id', 'edge_name', 'router_id']))

    # Also look for missing edges
    like_filters = {'router_id': vcns_const.BACKUP_ROUTER_PREFIX + "%"}
    rtr_bindings = nsxv_db.get_nsxv_router_bindings(edgeapi.context.session,
        like_filters=like_filters)
    plugin_nsx_missing = []

    for rtr_binding in rtr_bindings:
        if rtr_binding['edge_id'] not in backend_edge_ids:
            plugin_nsx_missing.append(
                {'edge_id': rtr_binding['edge_id'],
                 'router_id': rtr_binding['router_id'],
                 'db_status': rtr_binding['status']})

    LOG.info(formatters.output_formatter(
            constants.BACKUP_EDGES + ' missing from backend:',
            plugin_nsx_missing,
            ['edge_id', 'router_id', 'db_status']))


def nsx_fix_name_mismatch(resource, event, trigger, **kwargs):
    errmsg = ("Need to specify edge-id property. Add --property "
              "edge-id=<edge-id>")
    if not kwargs.get('property'):
        LOG.error("%s", errmsg)
        return
    properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
    edgeapi = utils.NeutronDbClient()
    edge_id = properties.get('edge-id')
    if not edge_id:
        LOG.error("%s", errmsg)
        return
    try:
        # edge[0] is response status code
        # edge[1] is response body
        edge = nsxv.get_edge(edge_id)[1]
    except exceptions.NeutronException as e:
        LOG.error("%s", str(e))
    else:
        if edge['name'].startswith('backup-'):

            rtr_binding = nsxv_db.get_nsxv_router_binding_by_edge(
                    edgeapi.context.session, edge['id'])

            if rtr_binding['router_id'] == edge['name']:
                LOG.error('Edge %s no mismatch with NSX', edge_id)
                return

            try:
                with locking.LockManager.get_lock(edge_id):
                    # Update edge at NSXv backend
                    if rtr_binding['router_id'].startswith('dhcp-'):
                        # Edge is a DHCP edge - just use router_id as name
                        edge['name'] = rtr_binding['router_id']
                    else:
                        # This is a router - if shared, prefix with 'shared-'
                        nsx_attr = (edgeapi.context.session.query(
                            nsxv_models.NsxvRouterExtAttributes).filter_by(
                                router_id=rtr_binding['router_id']).first())
                        if nsx_attr and nsx_attr['router_type'] == 'shared':
                            edge['name'] = ('shared-' + _uuid())[
                                           :vcns_const.EDGE_NAME_LEN]
                        elif (nsx_attr
                              and nsx_attr['router_type'] == 'exclusive'):
                            rtr_db = (edgeapi.context.session.query(
                                l3_db.Router).filter_by(
                                    id=rtr_binding['router_id']).first())
                            if rtr_db:
                                edge['name'] = (
                                    rtr_db['name'][
                                        :nsxv_constants.ROUTER_NAME_LENGTH -
                                        len(rtr_db['id'])] +
                                    '-' + rtr_db['id'])
                            else:
                                LOG.error(
                                    'No database entry for router id %s',
                                    rtr_binding['router_id'])

                        else:
                            LOG.error(
                                'Could not determine the name for '
                                'Edge %s', edge_id)
                            return

                    confirm = admin_utils.query_yes_no(
                        "Do you want to rename edge %s to %s" % (edge_id,
                                                                 edge['name']),
                        default="no")

                    if not confirm:
                        LOG.info("Edge rename aborted by user")
                        return
                    LOG.info("Edge rename started")
                    # remove some keys that will fail the NSX transaction
                    edge_utils.remove_irrelevant_keys_from_edge_request(edge)
                    try:
                        LOG.error("Update edge...")
                        nsxv.update_edge(edge_id, edge)
                    except Exception as e:
                        LOG.error("Update failed - %s", (e))
            except Exception as e:
                LOG.error("%s", str(e))
        else:
            LOG.error(
                'Edge %s has no backup prefix on NSX', edge_id)
            return

registry.subscribe(nsx_list_backup_edges,
                   constants.BACKUP_EDGES,
                   shell.Operations.LIST.value)
registry.subscribe(nsx_clean_backup_edge,
                   constants.BACKUP_EDGES,
                   shell.Operations.CLEAN.value)
registry.subscribe(nsx_clean_all_backup_edges,
                   constants.BACKUP_EDGES,
                   shell.Operations.CLEAN_ALL.value)
registry.subscribe(nsx_list_name_mismatches,
                   constants.BACKUP_EDGES,
                   shell.Operations.LIST_MISMATCHES.value)
registry.subscribe(nsx_fix_name_mismatch,
                   constants.BACKUP_EDGES,
                   shell.Operations.FIX_MISMATCH.value)
registry.subscribe(neutron_clean_backup_edge,
                   constants.BACKUP_EDGES,
                   shell.Operations.NEUTRON_CLEAN.value)



# Copyright 2016 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import registry
from oslo_log import log as logging

from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxv3.resources import utils
import vmware_nsx.shell.resources as shell
from vmware_nsxlib.v3 import nsx_constants

LOG = logging.getLogger(__name__)
neutron_client = utils.NeutronDbClient()


@admin_utils.output_header
def nsx_rate_limit_show(resource, event, trigger, **kwargs):
    """Show the current NSX rate limit."""

    nsxlib = utils.get_connected_nsxlib()
    if not nsxlib.feature_supported(nsx_constants.FEATURE_RATE_LIMIT):
        LOG.error("This utility is not available for NSX version %s",
                  nsxlib.get_version())
        return

    rate_limit = nsxlib.http_services.get_rate_limit()
    LOG.info("Current NSX rate limit is %s", rate_limit)


@admin_utils.output_header
def nsx_rate_limit_update(resource, event, trigger, **kwargs):
    """Set the NSX rate limit

    The default value is 40. 0 means no limit
    """
    nsxlib = utils.get_connected_nsxlib()
    if not nsxlib.feature_supported(nsx_constants.FEATURE_RATE_LIMIT):
        LOG.error("This utility is not available for NSX version %s",
                  nsxlib.get_version())
        return

    rate_limit = None
    if kwargs.get('property'):
        properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
        rate_limit = properties.get('value', None)
    if rate_limit is None:
        usage = ("nsxadmin -r rate-limit -o nsx-update "
                 "--property value=<new limit>")
        LOG.error("Missing parameters. Usage: %s", usage)
        return

    nsxlib.http_services.update_rate_limit(rate_limit)
    LOG.info("NSX rate limit was updated to %s", rate_limit)


registry.subscribe(nsx_rate_limit_show,
                   constants.RATE_LIMIT,
                   shell.Operations.SHOW.value)

registry.subscribe(nsx_rate_limit_update,
                   constants.RATE_LIMIT,
                   shell.Operations.NSX_UPDATE.value)



# Copyright (c) 2017 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from oslo_config import cfg
from oslo_utils import uuidutils

from neutron_lib import context
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory

from vmware_nsx.tests.unit.dvs import test_plugin as dvs_tests
from vmware_nsx.tests.unit.nsx_v import test_plugin as v_tests
from vmware_nsx.tests.unit.nsx_v3 import test_plugin as t_tests

PLUGIN_NAME = 'vmware_nsx.plugin.NsxTVDPlugin'
_uuid = uuidutils.generate_uuid


class NsxTVDPluginTestCase(v_tests.NsxVPluginV2TestCase,
                           t_tests.NsxV3PluginTestCaseMixin,
                           dvs_tests.NeutronSimpleDvsTestCase):

    def setUp(self,
              plugin=PLUGIN_NAME,
              ext_mgr=None,
              service_plugins=None):

        # set the default plugin
        if self.plugin_type:
            cfg.CONF.set_override('default_plugin', self.plugin_type,
                                  group="nsx_tvd")

        # set the default availability zones
        cfg.CONF.set_override('nsx_v_default_availability_zones',
                              ['default'],
                              group="nsx_tvd")
        cfg.CONF.set_override('nsx_v3_default_availability_zones',
                              ['defaultv3'],
                              group="nsx_tvd")

        super(NsxTVDPluginTestCase, self).setUp(
            plugin=plugin,
            ext_mgr=ext_mgr)
        self._project_id = _uuid()
        self.core_plugin = directory.get_plugin()

        # create a context with this tenant
        self.context = context.get_admin_context()
        self.context.tenant_id = self.project_id

        # create a default user for this plugin
        self.core_plugin.create_project_plugin_map(self.context,
                {'project_plugin_map': {'plugin': self.plugin_type,
                                        'project': self.project_id}})
        self.sub_plugin = self.core_plugin.get_plugin_by_type(self.plugin_type)

    @property
    def project_id(self):
        return self._project_id

    @property
    def plugin_type(self):
        pass

    def _test_plugin_initialized(self):
        self.assertTrue(self.core_plugin.is_tvd_plugin())
        self.assertIsNotNone(self.sub_plugin)

    def _test_call_create(self, obj_name, calls_count=1, project_id=None,
                          is_bulk=False):
        method_name = single_name = 'create_%s' % obj_name
        if is_bulk:
            method_name = method_name + '_bulk'
        func_to_call = getattr(self.core_plugin, method_name)
        if not project_id:
            project_id = self.project_id
        with mock.patch.object(self.sub_plugin, method_name) as sub_func,\
            mock.patch.object(self.sub_plugin, single_name) as single_func:
            if is_bulk:
                func_to_call(self.context,
                             {obj_name + 's': [{obj_name:
                                                {'tenant_id': project_id}}]})
            else:
                func_to_call(self.context,
                             {obj_name: {'tenant_id': project_id}})
            self.assertEqual(calls_count,
                             sub_func.call_count or single_func.call_count)

    def _test_call_create_with_net_id(self, obj_name, field_name='network_id',
                                      calls_count=1, is_bulk=False):
        method_name = 'create_%s' % obj_name
        if is_bulk:
            method_name = method_name + '_bulk'
        func_to_call = getattr(self.core_plugin, method_name)
        net_id = _uuid()

        with mock.patch.object(self.sub_plugin, method_name) as sub_func,\
            mock.patch.object(self.core_plugin, '_get_network',
                              return_value={'tenant_id': self.project_id}):
            if is_bulk:
                func_to_call(self.context,
                             {obj_name + 's': [{obj_name:
                                                {'tenant_id': self.project_id,
                                                 field_name: net_id}}]})
            else:
                func_to_call(self.context,
                             {obj_name: {'tenant_id': self.project_id,
                                         field_name: net_id}})
            self.assertEqual(calls_count, sub_func.call_count)

    def _test_call_delete(self, obj_name):
        method_name = 'delete_%s' % obj_name
        func_to_call = getattr(self.core_plugin, method_name)
        obj_id = _uuid()
        with mock.patch.object(self.sub_plugin, method_name) as sub_func,\
            mock.patch.object(self.core_plugin, '_get_%s' % obj_name,
                              return_value={'tenant_id': self.project_id}):
            func_to_call(self.context, obj_id)
            sub_func.assert_called_once()

    def _test_call_delete_with_net(self, obj_name, field_name='network_id'):
        method_name = 'delete_%s' % obj_name
        func_to_call = getattr(self.core_plugin, method_name)
        obj_id = _uuid()
        net_id = _uuid()
        with mock.patch.object(self.sub_plugin, method_name) as sub_func,\
            mock.patch.object(self.core_plugin, '_get_%s' % obj_name,
                              return_value={field_name: net_id}),\
            mock.patch.object(self.core_plugin, '_get_network',
                              return_value={'tenant_id': self.project_id}):
            func_to_call(self.context, obj_id)
            sub_func.assert_called_once()

    def _test_call_update(self, obj_name):
        method_name = 'update_%s' % obj_name
        func_to_call = getattr(self.core_plugin, method_name)
        obj_id = _uuid()
        with mock.patch.object(self.sub_plugin, method_name) as sub_func,\
            mock.patch.object(self.core_plugin, '_get_%s' % obj_name,
                              return_value={'tenant_id': self.project_id}):
            func_to_call(self.context, obj_id, {obj_name: {}})
            sub_func.assert_called_once()

    def _test_call_update_with_net(self, obj_name, field_name='network_id'):
        method_name = 'update_%s' % obj_name
        func_to_call = getattr(self.core_plugin, method_name)
        obj_id = _uuid()
        net_id = _uuid()
        with mock.patch.object(self.sub_plugin, method_name) as sub_func,\
            mock.patch.object(self.core_plugin, '_get_%s' % obj_name,
                              return_value={field_name: net_id}),\
            mock.patch.object(self.core_plugin, '_get_network',
                              return_value={'tenant_id': self.project_id}):
            func_to_call(self.context, obj_id, {obj_name: {}})
            sub_func.assert_called_once()

    def _test_call_get(self, obj_name):
        method_name = 'get_%s' % obj_name
        func_to_call = getattr(self.core_plugin, method_name)
        obj_id = _uuid()
        with mock.patch.object(self.sub_plugin, method_name) as sub_func,\
            mock.patch.object(self.core_plugin, '_get_%s' % obj_name,
                              return_value={'tenant_id': self.project_id}):
            func_to_call(self.context, obj_id)
            sub_func.assert_called_once()

    def _test_call_get_with_net(self, obj_name, field_name='network_id'):
        method_name = 'get_%s' % obj_name
        func_to_call = getattr(self.core_plugin, method_name)
        obj_id = _uuid()
        net_id = _uuid()
        with mock.patch.object(self.sub_plugin, method_name) as sub_func,\
            mock.patch.object(self.core_plugin, '_get_%s' % obj_name,
                              return_value={field_name: net_id}),\
            mock.patch.object(self.core_plugin, '_get_network',
                              return_value={'tenant_id': self.project_id}):
            func_to_call(self.context, obj_id)
            sub_func.assert_called_once()


class TestPluginWithDefaultPlugin(NsxTVDPluginTestCase):
    """Test TVD plugin with the NSX-T (default) sub plugin"""

    @property
    def plugin_type(self):
        return 'nsx-t'

    def test_plugin_initialized(self):
        self._test_plugin_initialized()

        # no unsupported extensions for the nsx_t plugin
        self.assertItemsEqual(
            ['router_type', 'router_size'],
            self.core_plugin._unsupported_fields[self.plugin_type]['router'])
        self.assertEqual(
            [],
            self.core_plugin._unsupported_fields[self.plugin_type]['port'])

    def test_create_network(self):
        self._test_call_create('network')

    def test_create_subnet(self):
        self._test_call_create_with_net_id('subnet')

    def test_create_port(self):
        self._test_call_create_with_net_id('port')

    def test_create_router(self):
        self._test_call_create('router')

    def test_create_floatingip(self):
        self._test_call_create_with_net_id(
            'floatingip', field_name='floating_network_id')

    def test_create_security_group(self):
        # plugin will be called twice because of the default sg
        self._test_call_create('security_group', calls_count=2)

    def test_create_security_group_rule(self):
        self._test_call_create('security_group_rule')

    def test_create_network_bulk(self):
        self._test_call_create('network', is_bulk=True)

    def test_create_subnet_bulk(self):
        self._test_call_create_with_net_id('subnet', is_bulk=True)

    def test_create_security_group_rule_bulk(self):
        self._test_call_create('security_group_rule', is_bulk=True)

    def test_delete_network(self):
        self._test_call_delete('network')

    def test_delete_subnet(self):
        self._test_call_delete_with_net('subnet')

    def test_delete_port(self):
        self._test_call_delete_with_net('port')

    def test_delete_router(self):
        self._test_call_delete('router')

    def test_delete_floatingip(self):
        self._test_call_delete_with_net(
            'floatingip', field_name='floating_network_id')

    def test_delete_security_group(self):
        self._test_call_delete('security_group')

    def test_update_network(self):
        self._test_call_update('network')

    def test_update_subnet(self):
        self._test_call_update_with_net('subnet')

    def test_update_port(self):
        self._test_call_update_with_net('port')

    def test_update_router(self):
        self._test_call_update('router')

    def test_update_floatingip(self):
        self._test_call_update_with_net(
            'floatingip', field_name='floating_network_id')

    def test_update_security_group(self):
        self._test_call_update('security_group')

    def test_unsupported_extensions(self):
        self.assertRaises(n_exc.InvalidInput,
            self.core_plugin.create_router,
            self.context,
            {'router': {'tenant_id': self.project_id,
                        'router_type': 'exclusive'}})

    def test_get_network(self):
        self._test_call_get('network')

    def test_get_subnet(self):
        self._test_call_get_with_net('subnet')

    def test_get_port(self):
        self._test_call_get_with_net('port')

    def test_get_router(self):
        self._test_call_get('router')

    def test_get_floatingip(self):
        self._test_call_get_with_net(
            'floatingip', field_name='floating_network_id')

    def test_get_security_group(self):
        self._test_call_get('security_group')

    def test_add_router_interface(self):
        rtr_id = _uuid()
        port_id = _uuid()
        net_id = _uuid()
        with mock.patch.object(self.sub_plugin,
                               'add_router_interface') as sub_func,\
            mock.patch.object(self.core_plugin, '_get_router',
                              return_value={'tenant_id': self.project_id}),\
            mock.patch.object(self.core_plugin, '_get_port',
                              return_value={'network_id': net_id}),\
            mock.patch.object(self.core_plugin, '_get_network',
                              return_value={'tenant_id': self.project_id}),\
            mock.patch.object(self.core_plugin, '_validate_interface_info',
                              return_value=(True, False)):
            self.core_plugin.add_router_interface(self.context, rtr_id,
                                                  {'port_id': port_id})
            sub_func.assert_called_once()

    def test_add_invalid_router_interface(self):
        # Test that the plugin prevents adding interface from one plugin
        # to a router of another plugin
        rtr_id = _uuid()
        port_id = _uuid()
        net_id = _uuid()
        another_tenant_id = _uuid()
        another_plugin = 'nsx-v' if self.plugin_type == 'nsx-t' else 'nsx-t'
        self.core_plugin.create_project_plugin_map(self.context,
                {'project_plugin_map': {'plugin': another_plugin,
                                        'project': another_tenant_id}})

        with mock.patch.object(self.core_plugin, '_get_router',
                               return_value={'tenant_id': self.project_id}),\
            mock.patch.object(self.core_plugin, '_get_port',
                              return_value={'network_id': net_id}),\
            mock.patch.object(self.core_plugin, '_get_network',
                              return_value={'tenant_id': another_tenant_id}),\
            mock.patch.object(self.core_plugin, '_validate_interface_info',
                              return_value=(True, False)):
            self.assertRaises(n_exc.InvalidInput,
                              self.core_plugin.add_router_interface,
                              self.context, rtr_id, {'port_id': port_id})

    def test_remove_router_interface(self):
        rtr_id = _uuid()
        with mock.patch.object(self.sub_plugin,
                               'remove_router_interface') as sub_func,\
            mock.patch.object(self.core_plugin, '_get_router',
                              return_value={'tenant_id': self.project_id}):
            self.core_plugin.remove_router_interface(self.context, rtr_id, {})
            sub_func.assert_called_once()

    def test_disassociate_floatingips(self):
        port_id = _uuid()
        net_id = _uuid()
        with mock.patch.object(self.sub_plugin,
                               'disassociate_floatingips') as sub_func,\
            mock.patch.object(self.core_plugin, '_get_port',
                              return_value={'network_id': net_id}),\
            mock.patch.object(self.core_plugin, '_get_network',
                              return_value={'tenant_id': self.project_id}):
            self.core_plugin.disassociate_floatingips(self.context, port_id)
            sub_func.assert_called_once()

    def test_new_user(self):
        project_id = _uuid()
        self._test_call_create('network', project_id=project_id)


class TestPluginWithNsxv(TestPluginWithDefaultPlugin):
    """Test TVD plugin with the NSX-V sub plugin"""

    @property
    def plugin_type(self):
        return 'nsx-v'

    def test_plugin_initialized(self):
        self._test_plugin_initialized()

        # no unsupported extensions for the nsx_v plugin
        self.assertEqual(
            [],
            self.core_plugin._unsupported_fields[self.plugin_type]['router'])
        self.assertEqual(
            [],
            self.core_plugin._unsupported_fields[self.plugin_type]['port'])

    def test_unsupported_extensions(self):
        self.skipTest('No unsupported extensions in this plugin')


class TestPluginWithDvs(TestPluginWithDefaultPlugin):
    """Test TVD plugin with the DVS sub plugin"""

    @property
    def plugin_type(self):
        return 'dvs'

    def test_plugin_initialized(self):
        self._test_plugin_initialized()

        # no unsupported extensions for the dvs plugin
        self.assertItemsEqual(
            ['mac_learning_enabled', 'provider_security_groups'],
            self.core_plugin._unsupported_fields[self.plugin_type]['port'])

    def test_unsupported_extensions(self):
        net_id = _uuid()
        with mock.patch.object(self.core_plugin, '_get_network',
                               return_value={'tenant_id': self.project_id}):
            self.assertRaises(n_exc.InvalidInput,
                self.core_plugin.create_port,
                self.context,
                {'port': {'tenant_id': self.project_id,
                          'network_id': net_id,
                          'mac_learning_enabled': True}})



#!/usr/bin/env python
# Copyright 2016 VMware Inc
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Purpose: Configure distributed firewall autodraft setting

Usage:
    python nsxv_fw_autodraft_setting.py --vsm-ip <nsx-manager-ip>
                                  --username <nsx-manager-username>
                                  --password <nsx-manager-password>
                                  [--autodraft-disable]
                                  [--autodraft-enable]
"""

import base64
import optparse

from oslo_serialization import jsonutils
import requests

requests.packages.urllib3.disable_warnings()


GLOBAL_CONFIG_URI = 'api/4.0/firewall/config/globalconfiguration'
AUTO_DRAFT_DISABLED = 'autoDraftDisabled'


class NSXClient(object):

    def __init__(self, host, username, password, *args, **kwargs):
        self._host = host
        self._username = username
        self._password = password

    def _get_headers(self, format):
        auth_cred = self._username + ":" + self._password
        auth = base64.b64encode(auth_cred)
        headers = {}
        headers['Authorization'] = "Basic %s" % auth
        headers['Content-Type'] = "application/%s" % format
        headers['Accept'] = "application/%s" % format
        return headers

    def _get_url(self, uri):
        return 'https://%s/%s' % (self._host, uri)

    def _get(self, format, uri):
        headers = self._get_headers(format)
        url = self._get_url(uri)
        response = requests.get(url, headers=headers,
                                verify=False)
        return response

    def _put(self, format, uri, data):
        headers = self._get_headers(format)
        url = self._get_url(uri)
        response = requests.put(url, headers=headers,
                                verify=False, data=data)
        return response

    def disable_autodraft(self):
        self._set_autodraft(True)

    def enable_autodraft(self):
        self._set_autodraft(False)

    def _get_global_config(self):
        resp = self._get('json', GLOBAL_CONFIG_URI)
        global_conf = jsonutils.loads(resp.text)
        return global_conf

    def _set_autodraft(self, disabled):
        global_conf = self._get_global_config()
        global_conf[AUTO_DRAFT_DISABLED] = disabled
        self._put('json', GLOBAL_CONFIG_URI, jsonutils.dumps(global_conf))


if __name__ == "__main__":
    parser = optparse.OptionParser()
    parser.add_option("--vsm-ip", dest="vsm_ip", help="NSX Manager IP address")
    parser.add_option("-u", "--username", default="admin", dest="username",
                      help="NSX Manager username")
    parser.add_option("-p", "--password", default="default", dest="password",
                      help="NSX Manager password")
    parser.add_option("--disable-autodraft", action="store_true",
                      default=False, dest="disabled",
                      help="Disable the autodraft setting for NSX "
                      "distributed firewal.")
    parser.add_option("--enable-autodraft", action="store_true",
                      default=False, dest="enabled",
                      help="Enable the autodraft setting for NSX "
                      "distributed firewal.")
    (options, args) = parser.parse_args()
    print("vsm-ip: %s" % options.vsm_ip)
    print("username: %s" % options.username)
    print("password: %s" % options.password)

    if options.disabled and options.enabled:
        print("Please provide only one of the options: --disable-autodraft, "
              "--enable-autodraft.")

    nsx_client = NSXClient(options.vsm_ip, options.username,
                           options.password)
    if options.disabled:
        print("Disabling autodraft settings:")
        nsx_client.disable_autodraft()
        print("Autodraft is now disabled.")
    if options.enabled:
        print("Enabling autodraft settings:")
        nsx_client.enable_autodraft()
        print("Autodraft is now enabled.")



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests.unit.extensions import test_portsecurity as psec
from vmware_nsx.common import sync
from vmware_nsx.tests import unit as vmware
from vmware_nsx.tests.unit.nsx_mh.apiclient import fake
from vmware_nsx.tests.unit.nsx_v3 import test_constants as v3_constants
from vmware_nsx.tests.unit.nsx_v3 import test_plugin as test_nsxv3
from vmware_nsx.tests.unit import test_utils


class PortSecurityTestCaseNSXv2(psec.PortSecurityDBTestCase):

    def setUp(self):
        test_utils.override_nsx_ini_test()
        # mock api client
        self.fc = fake.FakeClient(vmware.STUBS_PATH)
        self.mock_nsx = mock.patch(vmware.NSXAPI_NAME, autospec=True)
        instance = self.mock_nsx.start()
        instance.return_value.login.return_value = "the_cookie"
        # Avoid runs of the synchronizer looping call
        patch_sync = mock.patch.object(sync, '_start_loopingcall')
        patch_sync.start()

        instance.return_value.request.side_effect = self.fc.fake_request
        super(PortSecurityTestCaseNSXv2, self).setUp(vmware.PLUGIN_NAME)
        self.addCleanup(self.fc.reset_all)
        self.addCleanup(self.mock_nsx.stop)
        self.addCleanup(patch_sync.stop)


class TestPortSecurityNSXv2(PortSecurityTestCaseNSXv2, psec.TestPortSecurity):
        pass


class TestPortSecurityNSXv3(psec.TestPortSecurity,
                            test_nsxv3.NsxV3PluginTestCaseMixin):
    def setUp(self, plugin=v3_constants.PLUGIN_NAME):
        super(TestPortSecurityNSXv3, self).setUp(plugin=plugin)



# Copyright 2012 VMware, Inc.
#
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from six.moves import http_client as httplib

from vmware_nsx._i18n import _


def ctrl_conn_to_str(conn):
    """Returns a string representing a connection URL to the controller."""
    if isinstance(conn, httplib.HTTPSConnection):
        proto = "https://"
    elif isinstance(conn, httplib.HTTPConnection):
        proto = "http://"
    else:
        raise TypeError(_('Invalid connection type: %s') % type(conn))
    return "%s%s:%s" % (proto, conn.host, conn.port)



# Copyright 2013 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from sqlalchemy.orm import exc

from neutron.db import _model_query as model_query
from neutron.db import _resource_extend as resource_extend
from neutron.db import _utils as db_utils
from neutron.db import api as db_api

from neutron_lib.api.definitions import port as port_def

from oslo_log import log as logging

from vmware_nsx.db import nsx_models
from vmware_nsx.extensions import maclearning as mac

LOG = logging.getLogger(__name__)


@resource_extend.has_resource_extenders
class MacLearningDbMixin(object):
    """Mixin class for mac learning."""

    def _make_mac_learning_state_dict(self, port, fields=None):
        res = {'port_id': port['port_id'],
               mac.MAC_LEARNING: port[mac.MAC_LEARNING]}
        return db_utils.resource_fields(res, fields)

    @staticmethod
    @resource_extend.extends([port_def.COLLECTION_NAME])
    def _extend_port_mac_learning_state(port_res, port_db):
        state = port_db.mac_learning_state
        if state and state.mac_learning_enabled:
            port_res[mac.MAC_LEARNING] = state.mac_learning_enabled

    def _update_mac_learning_state(self, context, port_id, enabled):
        try:
            query = model_query.query_with_hooks(
                context, nsx_models.MacLearningState)
            state = query.filter(
                nsx_models.MacLearningState.port_id == port_id).one()
            state.update({mac.MAC_LEARNING: enabled})
        except exc.NoResultFound:
            self._create_mac_learning_state(context,
                                            {'id': port_id,
                                             mac.MAC_LEARNING: enabled})

    def _create_mac_learning_state(self, context, port):
        with db_api.context_manager.writer.using(context):
            enabled = port[mac.MAC_LEARNING]
            state = nsx_models.MacLearningState(
                port_id=port['id'],
                mac_learning_enabled=enabled)
            context.session.add(state)
        return self._make_mac_learning_state_dict(state)

    def get_mac_learning_state(self, context, port_id):
        try:
            query = model_query.query_with_hooks(
                context, nsx_models.MacLearningState)
            state = query.filter(
                nsx_models.MacLearningState.port_id == port_id).one()
            return state.mac_learning_enabled
        except exc.NoResultFound:
            return None



# Copyright (c) 2012 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg


def override_nsx_ini_test():
    cfg.CONF.set_override("default_tz_uuid", "fake_tz_uuid")
    cfg.CONF.set_override("nsx_controllers", ["fake1", "fake_2"])
    cfg.CONF.set_override("nsx_user", "foo")
    cfg.CONF.set_override("nsx_password", "bar")
    cfg.CONF.set_override("default_l3_gw_service_uuid", "whatever")
    cfg.CONF.set_override("default_l2_gw_service_uuid", "whatever")
    cfg.CONF.set_override("manager_uri", "https://fake_manager",
                          group="nsxv")
    cfg.CONF.set_override("user", "fake_user", group="nsxv")
    cfg.CONF.set_override("password", "fake_password", group="nsxv")
    cfg.CONF.set_override("vdn_scope_id", "fake_vdn_scope_id",
                          group="nsxv")
    cfg.CONF.set_override("dvs_id", "fake_dvs_id", group="nsxv")


def override_nsx_ini_full_test():
    cfg.CONF.set_override("default_tz_uuid", "fake_tz_uuid")
    cfg.CONF.set_override("nsx_controllers", ["fake1", "fake_2"])
    cfg.CONF.set_override("nsx_user", "foo")
    cfg.CONF.set_override("nsx_password", "bar")
    cfg.CONF.set_override("default_l3_gw_service_uuid", "whatever")
    cfg.CONF.set_override("default_l2_gw_service_uuid", "whatever")
    cfg.CONF.set_override("nsx_default_interface_name", "whatever")
    cfg.CONF.set_override("http_timeout", 13)
    cfg.CONF.set_override("redirects", 12)
    cfg.CONF.set_override("retries", "11")



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
import webob.exc

from neutron_lib.api.definitions import multiprovidernet as mpnet_apidef
from neutron_lib.api.definitions import provider_net as pnet
from vmware_nsx.tests import unit as vmware
from vmware_nsx.tests.unit.nsx_mh import test_plugin as test_nsx_plugin


class TestProvidernet(test_nsx_plugin.NsxPluginV2TestCase):

    def test_create_delete_provider_network_default_physical_net(self):
        data = {'network': {'name': 'net1',
                            'admin_state_up': True,
                            'tenant_id': 'admin',
                            pnet.NETWORK_TYPE: 'vlan',
                            pnet.SEGMENTATION_ID: 411}}
        network_req = self.new_create_request('networks', data, self.fmt)
        net = self.deserialize(self.fmt, network_req.get_response(self.api))
        self.assertEqual(net['network'][pnet.NETWORK_TYPE], 'vlan')
        self.assertEqual(net['network'][pnet.SEGMENTATION_ID], 411)
        req = self.new_delete_request('networks', net['network']['id'])
        res = req.get_response(self.api)
        self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)

    def test_create_provider_network(self):
        data = {'network': {'name': 'net1',
                            'admin_state_up': True,
                            'tenant_id': 'admin',
                            pnet.NETWORK_TYPE: 'vlan',
                            pnet.SEGMENTATION_ID: 411,
                            pnet.PHYSICAL_NETWORK: 'physnet1'}}
        network_req = self.new_create_request('networks', data, self.fmt)
        net = self.deserialize(self.fmt, network_req.get_response(self.api))
        self.assertEqual(net['network'][pnet.NETWORK_TYPE], 'vlan')
        self.assertEqual(net['network'][pnet.SEGMENTATION_ID], 411)
        self.assertEqual(net['network'][pnet.PHYSICAL_NETWORK], 'physnet1')

        # Test that we can create another provider network using the same
        # vlan_id on another physical network.
        data['network'][pnet.PHYSICAL_NETWORK] = 'physnet2'
        network_req = self.new_create_request('networks', data, self.fmt)
        net = self.deserialize(self.fmt, network_req.get_response(self.api))
        self.assertEqual(net['network'][pnet.NETWORK_TYPE], 'vlan')
        self.assertEqual(net['network'][pnet.SEGMENTATION_ID], 411)
        self.assertEqual(net['network'][pnet.PHYSICAL_NETWORK], 'physnet2')


class TestMultiProviderNetworks(test_nsx_plugin.NsxPluginV2TestCase):

    def setUp(self, plugin=None):
        cfg.CONF.set_override('api_extensions_path', vmware.NSXEXT_PATH)
        super(TestMultiProviderNetworks, self).setUp()

    def test_create_network_provider(self):
        data = {'network': {'name': 'net1',
                            pnet.NETWORK_TYPE: 'vlan',
                            pnet.PHYSICAL_NETWORK: 'physnet1',
                            pnet.SEGMENTATION_ID: 1,
                            'tenant_id': 'tenant_one'}}
        network_req = self.new_create_request('networks', data)
        network = self.deserialize(self.fmt,
                                   network_req.get_response(self.api))
        self.assertEqual(network['network'][pnet.NETWORK_TYPE], 'vlan')
        self.assertEqual(network['network'][pnet.PHYSICAL_NETWORK], 'physnet1')
        self.assertEqual(network['network'][pnet.SEGMENTATION_ID], 1)
        self.assertNotIn(mpnet_apidef.SEGMENTS, network['network'])

    def test_create_network_provider_flat(self):
        data = {'network': {'name': 'net1',
                            pnet.NETWORK_TYPE: 'flat',
                            pnet.PHYSICAL_NETWORK: 'physnet1',
                            'tenant_id': 'tenant_one'}}
        network_req = self.new_create_request('networks', data)
        network = self.deserialize(self.fmt,
                                   network_req.get_response(self.api))
        self.assertEqual('flat', network['network'][pnet.NETWORK_TYPE])
        self.assertEqual('physnet1', network['network'][pnet.PHYSICAL_NETWORK])
        self.assertEqual(0, network['network'][pnet.SEGMENTATION_ID])
        self.assertNotIn(mpnet_apidef.SEGMENTS, network['network'])

    def test_create_network_single_multiple_provider(self):
        data = {'network': {'name': 'net1',
                            mpnet_apidef.SEGMENTS:
                            [{pnet.NETWORK_TYPE: 'vlan',
                              pnet.PHYSICAL_NETWORK: 'physnet1',
                              pnet.SEGMENTATION_ID: 1}],
                            'tenant_id': 'tenant_one'}}
        net_req = self.new_create_request('networks', data)
        network = self.deserialize(self.fmt, net_req.get_response(self.api))
        for provider_field in [pnet.NETWORK_TYPE, pnet.PHYSICAL_NETWORK,
                               pnet.SEGMENTATION_ID]:
            self.assertNotIn(provider_field, network['network'])
        tz = network['network'][mpnet_apidef.SEGMENTS][0]
        self.assertEqual(tz[pnet.NETWORK_TYPE], 'vlan')
        self.assertEqual(tz[pnet.PHYSICAL_NETWORK], 'physnet1')
        self.assertEqual(tz[pnet.SEGMENTATION_ID], 1)

        # Tests get_network()
        net_req = self.new_show_request('networks', network['network']['id'])
        network = self.deserialize(self.fmt, net_req.get_response(self.api))
        tz = network['network'][mpnet_apidef.SEGMENTS][0]
        self.assertEqual(tz[pnet.NETWORK_TYPE], 'vlan')
        self.assertEqual(tz[pnet.PHYSICAL_NETWORK], 'physnet1')
        self.assertEqual(tz[pnet.SEGMENTATION_ID], 1)

    def test_create_network_multprovider(self):
        data = {'network': {'name': 'net1',
                            mpnet_apidef.SEGMENTS:
                            [{pnet.NETWORK_TYPE: 'vlan',
                              pnet.PHYSICAL_NETWORK: 'physnet1',
                              pnet.SEGMENTATION_ID: 1},
                             {pnet.NETWORK_TYPE: 'stt',
                              pnet.PHYSICAL_NETWORK: 'physnet1'}],
                            'tenant_id': 'tenant_one'}}
        network_req = self.new_create_request('networks', data)
        network = self.deserialize(self.fmt,
                                   network_req.get_response(self.api))
        tz = network['network'][mpnet_apidef.SEGMENTS]
        for tz in data['network'][mpnet_apidef.SEGMENTS]:
            for field in [pnet.NETWORK_TYPE, pnet.PHYSICAL_NETWORK,
                          pnet.SEGMENTATION_ID]:
                self.assertEqual(tz.get(field), tz.get(field))

        # Tests get_network()
        net_req = self.new_show_request('networks', network['network']['id'])
        network = self.deserialize(self.fmt, net_req.get_response(self.api))
        tz = network['network'][mpnet_apidef.SEGMENTS]
        for tz in data['network'][mpnet_apidef.SEGMENTS]:
            for field in [pnet.NETWORK_TYPE, pnet.PHYSICAL_NETWORK,
                          pnet.SEGMENTATION_ID]:
                self.assertEqual(tz.get(field), tz.get(field))

    def test_create_network_with_provider_and_multiprovider_fail(self):
        data = {'network': {'name': 'net1',
                            mpnet_apidef.SEGMENTS:
                            [{pnet.NETWORK_TYPE: 'vlan',
                              pnet.PHYSICAL_NETWORK: 'physnet1',
                              pnet.SEGMENTATION_ID: 1}],
                            pnet.NETWORK_TYPE: 'vlan',
                            pnet.PHYSICAL_NETWORK: 'physnet1',
                            pnet.SEGMENTATION_ID: 1,
                            'tenant_id': 'tenant_one'}}

        network_req = self.new_create_request('networks', data)
        res = network_req.get_response(self.api)
        self.assertEqual(res.status_int, 400)

    def test_create_network_duplicate_segments(self):
        data = {'network': {'name': 'net1',
                            mpnet_apidef.SEGMENTS:
                            [{pnet.NETWORK_TYPE: 'vlan',
                              pnet.PHYSICAL_NETWORK: 'physnet1',
                              pnet.SEGMENTATION_ID: 1},
                             {pnet.NETWORK_TYPE: 'vlan',
                              pnet.PHYSICAL_NETWORK: 'physnet1',
                              pnet.SEGMENTATION_ID: 1}],
                            'tenant_id': 'tenant_one'}}
        network_req = self.new_create_request('networks', data)
        res = network_req.get_response(self.api)
        self.assertEqual(res.status_int, 400)



# Copyright 2016 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import xml.etree.ElementTree as et

from neutron.db import api as db_api
from neutron.db.models import securitygroup as sg_models
from neutron.db import models_v2
from neutron.db import securitygroups_db
from neutron.extensions import securitygroup as ext_sg
from neutron_lib.callbacks import registry
from neutron_lib import context as n_context
from oslo_log import log as logging

from vmware_nsx.common import utils as com_utils
from vmware_nsx.db import db as nsx_db
from vmware_nsx.db import extended_security_group as extended_secgroup
from vmware_nsx.db import extended_security_group_rule as extend_sg_rule
from vmware_nsx.db import nsx_models
from vmware_nsx.db import nsxv_db
from vmware_nsx.db import nsxv_models
from vmware_nsx.extensions import securitygrouppolicy as sg_policy
from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import formatters
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxv.resources import utils
from vmware_nsx.shell import resources as shell


LOG = logging.getLogger(__name__)


class NeutronSecurityGroupDB(
    utils.NeutronDbClient,
    securitygroups_db.SecurityGroupDbMixin,
    extended_secgroup.ExtendedSecurityGroupPropertiesMixin,
    extend_sg_rule.ExtendedSecurityGroupRuleMixin):

    def __init__(self):
        super(NeutronSecurityGroupDB, self)
        # FIXME(roeyc): context is already defined in NeutrondDbClient
        self.context = n_context.get_admin_context()

    def get_security_groups_mappings(self):
        q = self.context.session.query(
            sg_models.SecurityGroup.name,
            sg_models.SecurityGroup.id,
            nsxv_models.NsxvSecurityGroupSectionMapping.ip_section_id,
            nsx_models.NeutronNsxSecurityGroupMapping.nsx_id).join(
                nsxv_models.NsxvSecurityGroupSectionMapping,
                nsx_models.NeutronNsxSecurityGroupMapping).all()
        sg_mappings = [{'name': mapp.name,
                        'id': mapp.id,
                        'section-uri': mapp.ip_section_id,
                        'nsx-securitygroup-id': mapp.nsx_id}
                       for mapp in q]
        return sg_mappings

    def get_security_group(self, sg_id):
        return super(NeutronSecurityGroupDB, self).get_security_group(
            self.context, sg_id)

    def get_security_groups(self):
        filters = utils.get_plugin_filters(self.context)
        return super(NeutronSecurityGroupDB,
                     self).get_security_groups(self.context, filters=filters)

    def get_security_group_id_by_section_id(self, section_id):
        section_url = ("/api/4.0/firewall/globalroot-0/config/layer3sections"
                       "/%s" % section_id)
        q = self.context.session.query(
            nsxv_models.NsxvSecurityGroupSectionMapping).filter_by(
            ip_section_id=section_url).all()
        if q:
            return q[0].neutron_id

    def _is_provider_section(self, section_id):
        # look for this section id in the nsx_db, and get the security group
        sg_id = self.get_security_group_id_by_section_id(section_id)
        if sg_id:
            # Check in the DB if this is a provider SG
            return self._is_provider_security_group(self.context, sg_id)
        return False

    def delete_security_group_section_mapping(self, sg_id):
        with db_api.context_manager.writer.using(self.context):
            fw_mapping = self.context.session.query(
                nsxv_models.NsxvSecurityGroupSectionMapping).filter_by(
                    neutron_id=sg_id).one_or_none()
            if fw_mapping:
                self.context.session.delete(fw_mapping)

    def delete_security_group_backend_mapping(self, sg_id):
        with db_api.context_manager.writer.using(self.context):
            sg_mapping = self.context.session.query(
                nsx_models.NeutronNsxSecurityGroupMapping).filter_by(
                    neutron_id=sg_id).one_or_none()
            if sg_mapping:
                self.context.session.delete(sg_mapping)

    def get_vnics_in_security_group(self, security_group_id):
        with utils.NsxVPluginWrapper() as plugin:
            vnics = []
            query = self.context.session.query(
                models_v2.Port.id, models_v2.Port.device_id
            ).join(sg_models.SecurityGroupPortBinding).filter_by(
                security_group_id=security_group_id).all()
            for p in query:
                vnic_index = plugin._get_port_vnic_index(self.context, p.id)
                vnic_id = plugin._get_port_vnic_id(vnic_index, p.device_id)
                vnics.append(vnic_id)
            return vnics


class NsxFirewallAPI(object):
    def __init__(self):
        self.vcns = utils.get_nsxv_client()

    def list_security_groups(self):
        h, secgroups = self.vcns.list_security_groups()
        if not secgroups:
            return []
        root = et.fromstring(secgroups)
        secgroups = []
        for sg in root.iter('securitygroup'):
            sg_id = sg.find('objectId').text
            # This specific security-group is not relevant to the plugin
            if sg_id == 'securitygroup-1':
                continue
            secgroups.append({'name': sg.find('name').text,
                              'id': sg_id})
        return secgroups

    def list_fw_sections(self):
        h, firewall_config = self.vcns.get_dfw_config()
        if not firewall_config:
            return []
        root = com_utils.normalize_xml(firewall_config)
        sections = []
        for sec in root.iter('section'):
            sec_id = sec.attrib['id']
            # Don't show NSX default sections, which are not relevant to OS.
            if sec_id in ['1001', '1002', '1003']:
                continue
            sections.append({'name': sec.attrib['name'],
                             'id': sec_id})
        return sections

    def reorder_fw_sections(self):
        # read all the sections
        h, firewall_config = self.vcns.get_dfw_config()
        if not firewall_config:
            LOG.info("No firewall sections were found.")
            return

        root = com_utils.normalize_xml(firewall_config)

        for child in root:
            if str(child.tag) == 'layer3Sections':
                # go over the L3 sections and reorder them.
                # The correct order should be:
                # 1. OS provider security groups
                # 2. service composer policies
                # 3. regular OS security groups
                sections = list(child.iter('section'))
                provider_sections = []
                regular_sections = []
                policy_sections = []

                for sec in sections:
                    if sec.attrib.get('managedBy') == 'NSX Service Composer':
                        policy_sections.append(sec)
                    else:
                        if neutron_sg._is_provider_section(
                            sec.attrib.get('id')):
                            provider_sections.append(sec)
                        else:
                            regular_sections.append(sec)
                    child.remove(sec)

                if not policy_sections and not provider_sections:
                    LOG.info("No need to reorder the firewall sections.")
                    return

                # reorder the sections
                reordered_sections = (provider_sections +
                                      policy_sections +
                                      regular_sections)
                child.extend(reordered_sections)

                # update the new order of sections in the backend
                self.vcns.update_dfw_config(et.tostring(root), h)
                LOG.info("L3 Firewall sections were reordered.")


neutron_sg = NeutronSecurityGroupDB()
nsxv_firewall = NsxFirewallAPI()


def _log_info(resource, data, attrs=['name', 'id']):
    LOG.info(formatters.output_formatter(resource, data, attrs))


@admin_utils.list_handler(constants.SECURITY_GROUPS)
@admin_utils.output_header
def neutron_list_security_groups_mappings(resource, event, trigger, **kwargs):
    sg_mappings = neutron_sg.get_security_groups_mappings()
    _log_info(constants.SECURITY_GROUPS,
              sg_mappings,
              attrs=['name', 'id', 'section-uri', 'nsx-securitygroup-id'])
    return bool(sg_mappings)


@admin_utils.list_handler(constants.FIREWALL_SECTIONS)
@admin_utils.output_header
def nsx_list_dfw_sections(resource, event, trigger, **kwargs):
    fw_sections = nsxv_firewall.list_fw_sections()
    _log_info(constants.FIREWALL_SECTIONS, fw_sections)
    return bool(fw_sections)


@admin_utils.list_handler(constants.FIREWALL_NSX_GROUPS)
@admin_utils.output_header
def nsx_list_security_groups(resource, event, trigger, **kwargs):
    nsx_secgroups = nsxv_firewall.list_security_groups()
    _log_info(constants.FIREWALL_NSX_GROUPS, nsx_secgroups)
    return bool(nsx_secgroups)


def _find_missing_security_groups():
    nsx_secgroups = nsxv_firewall.list_security_groups()
    sg_mappings = neutron_sg.get_security_groups_mappings()
    missing_secgroups = {}
    for sg_db in sg_mappings:
        for nsx_sg in nsx_secgroups:
            if nsx_sg['id'] == sg_db['nsx-securitygroup-id']:
                break
        else:
            missing_secgroups[sg_db['id']] = sg_db
    return missing_secgroups


@admin_utils.list_mismatches_handler(constants.FIREWALL_NSX_GROUPS)
@admin_utils.output_header
def list_missing_security_groups(resource, event, trigger, **kwargs):
    sgs_with_missing_nsx_group = _find_missing_security_groups()
    missing_securitgroups_info = [
        {'securitygroup-name': sg['name'],
         'securitygroup-id': sg['id'],
         'nsx-securitygroup-id':
         sg['nsx-securitygroup-id']}
        for sg in sgs_with_missing_nsx_group.values()]
    _log_info(constants.FIREWALL_NSX_GROUPS, missing_securitgroups_info,
              attrs=['securitygroup-name', 'securitygroup-id',
                     'nsx-securitygroup-id'])
    return bool(missing_securitgroups_info)


def _find_missing_sections():
    fw_sections = nsxv_firewall.list_fw_sections()
    sg_mappings = neutron_sg.get_security_groups_mappings()
    missing_sections = {}
    for sg_db in sg_mappings:
        for fw_section in fw_sections:
            if fw_section['id'] == sg_db.get('section-uri', '').split('/')[-1]:
                break
        else:
            missing_sections[sg_db['id']] = sg_db
    return missing_sections


@admin_utils.list_mismatches_handler(constants.FIREWALL_SECTIONS)
@admin_utils.output_header
def list_missing_firewall_sections(resource, event, trigger, **kwargs):
    sgs_with_missing_section = _find_missing_sections()
    missing_sections_info = [{'securitygroup-name': sg['name'],
                              'securitygroup-id': sg['id'],
                              'section-id': sg['section-uri']}
                             for sg in sgs_with_missing_section.values()]
    _log_info(constants.FIREWALL_SECTIONS, missing_sections_info,
              attrs=['securitygroup-name', 'securitygroup-id', 'section-uri'])
    return bool(missing_sections_info)


@admin_utils.list_mismatches_handler(constants.FIREWALL_SECTIONS)
@admin_utils.output_header
def reorder_firewall_sections(resource, event, trigger, **kwargs):
    nsxv_firewall.reorder_fw_sections()


@admin_utils.fix_mismatches_handler(constants.SECURITY_GROUPS)
@admin_utils.output_header
def fix_security_groups(resource, event, trigger, **kwargs):
    context_ = n_context.get_admin_context()
    sgs_with_missing_section = _find_missing_sections()
    sgs_with_missing_nsx_group = _find_missing_security_groups()
    if not sgs_with_missing_section and not sgs_with_missing_nsx_group:
        # no mismatches
        return

    with utils.NsxVPluginWrapper() as plugin:
        # If only the fw section is missing then create it.
        for sg_id in (set(sgs_with_missing_section.keys()) -
                      set(sgs_with_missing_nsx_group.keys())):
            neutron_sg.delete_security_group_section_mapping(sg_id)
            secgroup = plugin.get_security_group(context_, sg_id)
            plugin._create_fw_section_for_security_group(
                context_, secgroup,
                sgs_with_missing_section[sg_id]['nsx-securitygroup-id'])

        # If nsx security-group is missing then create both nsx security-group
        # and a new fw section (remove old one).
        for sg_id, sg in sgs_with_missing_nsx_group.items():
            secgroup = plugin.get_security_group(context_, sg_id)
            if sg_id not in sgs_with_missing_section:
                plugin._delete_section(sg['section-uri'])
            neutron_sg.delete_security_group_section_mapping(sg_id)
            neutron_sg.delete_security_group_backend_mapping(sg_id)
            plugin._process_security_group_create_backend_resources(context_,
                                                                    secgroup)
            nsx_id = nsx_db.get_nsx_security_group_id(context_.session, sg_id,
                                                      moref=False)
            for vnic_id in neutron_sg.get_vnics_in_security_group(sg_id):
                plugin._add_member_to_security_group(nsx_id, vnic_id)


@admin_utils.output_header
def migrate_sg_to_policy(resource, event, trigger, **kwargs):
    """Change the mode of a security group from rules to NSX policy"""
    if not kwargs.get('property'):
        LOG.error("Need to specify security-group-id and policy-id "
                  "parameters")
        return

    # input validation
    properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
    sg_id = properties.get('security-group-id')
    if not sg_id:
        LOG.error("Need to specify security-group-id parameter")
        return
    policy_id = properties.get('policy-id')
    if not policy_id:
        LOG.error("Need to specify policy-id parameter")
        return

    # validate that the security group exist and contains rules and no policy
    context_ = n_context.get_admin_context()
    with utils.NsxVPluginWrapper() as plugin:
        try:
            secgroup = plugin.get_security_group(context_, sg_id)
        except ext_sg.SecurityGroupNotFound:
            LOG.error("Security group %s was not found", sg_id)
            return
        if secgroup.get('policy'):
            LOG.error("Security group %s already uses a policy", sg_id)
            return

        # validate that the policy exists
        if not plugin.nsx_v.vcns.validate_inventory(policy_id):
            LOG.error("NSX policy %s was not found", policy_id)
            return

        # get the nsx id from the backend
        nsx_sg_id = nsx_db.get_nsx_security_group_id(context_.session, sg_id,
                                                     moref=True)
        if not nsx_sg_id:
            LOG.error("Did not find security groups %s neutron ID", sg_id)
            return

        # Delete the rules from the security group
        LOG.info("Deleting the rules of security group: %s", sg_id)
        for rule in secgroup.get('security_group_rules', []):
            try:
                plugin.delete_security_group_rule(context_, rule['id'])
            except Exception as e:
                LOG.warning("Failed to delete rule %(r)s from security "
                            "group %(sg)s: %(e)s",
                            {'r': rule['id'], 'sg': sg_id, 'e': e})
                # continue anyway

        # Delete the security group FW section
        LOG.info("Deleting the section of security group: %s", sg_id)
        try:
            section_uri = plugin._get_section_uri(context_.session, sg_id)
            plugin._delete_section(section_uri)
            nsxv_db.delete_neutron_nsx_section_mapping(
                context_.session, sg_id)
        except Exception as e:
            LOG.warning("Failed to delete firewall section of security "
                        "group %(sg)s: %(e)s",
                        {'sg': sg_id, 'e': e})
            # continue anyway

        # bind this security group to the policy in the backend and DB
        LOG.info("Binding the NSX security group %(nsx)s to policy "
                 "%(pol)s",
                 {'nsx': nsx_sg_id, 'pol': policy_id})
        plugin._update_nsx_security_group_policies(
            policy_id, None, nsx_sg_id)
        with context_.session.begin(subtransactions=True):
            prop = context_.session.query(
                extended_secgroup.NsxExtendedSecurityGroupProperties).\
                filter_by(security_group_id=sg_id).one()
            prop[sg_policy.POLICY] = policy_id
        LOG.info("Done.")


@admin_utils.output_header
def firewall_update_cluster_default_fw_section(resource, event, trigger,
                                               **kwargs):
    with utils.NsxVPluginWrapper() as plugin:
        plugin._create_cluster_default_fw_section()
        LOG.info("Cluster default FW section updated.")


registry.subscribe(migrate_sg_to_policy,
                   constants.SECURITY_GROUPS,
                   shell.Operations.MIGRATE_TO_POLICY.value)

registry.subscribe(reorder_firewall_sections,
                   constants.FIREWALL_SECTIONS,
                   shell.Operations.NSX_REORDER.value)

registry.subscribe(fix_security_groups,
                   constants.FIREWALL_SECTIONS,
                   shell.Operations.NSX_UPDATE.value)

registry.subscribe(firewall_update_cluster_default_fw_section,
                   constants.FIREWALL_SECTIONS,
                   shell.Operations.NSX_UPDATE.value)



# Copyright 2014 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron import version
from neutron_lib import exceptions as exception
from oslo_log import log
from oslo_serialization import jsonutils
import six

from vmware_nsx._i18n import _
from vmware_nsx.api_client import exception as api_exc
from vmware_nsx.common import exceptions as nsx_exc

HTTP_GET = "GET"
HTTP_POST = "POST"
HTTP_DELETE = "DELETE"
HTTP_PUT = "PUT"
# Prefix to be used for all NSX API calls
URI_PREFIX = "/ws.v1"
NEUTRON_VERSION = version.version_info.release_string()

LOG = log.getLogger(__name__)


def _build_uri_path(resource,
                    resource_id=None,
                    parent_resource_id=None,
                    fields=None,
                    relations=None,
                    filters=None,
                    types=None,
                    is_attachment=False,
                    extra_action=None):
    resources = resource.split('/')
    res_path = resources[0]
    if resource_id:
        res_path += "/%s" % resource_id
    if len(resources) > 1:
        # There is also a parent resource to account for in the uri
        res_path = "%s/%s/%s" % (resources[1],
                                 parent_resource_id,
                                 res_path)
    if is_attachment:
        res_path = "%s/attachment" % res_path
    elif extra_action:
        res_path = "%s/%s" % (res_path, extra_action)
    params = []
    params.append(fields and "fields=%s" % fields)
    params.append(relations and "relations=%s" % relations)
    params.append(types and "types=%s" % types)
    if filters:
        sorted_filters = [
            '%s=%s' % (k, filters[k]) for k in sorted(filters.keys())
        ]
        params.extend(sorted_filters)
    uri_path = "%s/%s" % (URI_PREFIX, res_path)
    non_empty_params = [x for x in params if x is not None]
    if non_empty_params:
        query_string = '&'.join(non_empty_params)
        if query_string:
            uri_path += "?%s" % query_string
    return uri_path


def format_exception(etype, e, exception_locals):
    """Consistent formatting for exceptions.

    :param etype: a string describing the exception type.
    :param e: the exception.
    :param execption_locals: calling context local variable dict.
    :returns: a formatted string.
    """
    msg = [_("Error. %(type)s exception: %(exc)s.") %
           {'type': etype, 'exc': e}]
    l = dict((k, v) for k, v in six.iteritems(exception_locals)
             if k != 'request')
    msg.append(_("locals=[%s]") % str(l))
    return ' '.join(msg)


def do_request(*args, **kwargs):
    """Issue a request to the cluster specified in kwargs.

    :param args: a list of positional arguments.
    :param kwargs: a list of keyworkds arguments.
    :returns: the result of the operation loaded into a python
        object or None.
    """
    cluster = kwargs["cluster"]
    try:
        res = cluster.api_client.request(*args)
        if res:
            return jsonutils.loads(res)
    except api_exc.ResourceNotFound:
        raise exception.NotFound()
    except api_exc.ReadOnlyMode:
        raise nsx_exc.MaintenanceInProgress()


def get_single_query_page(path, cluster, page_cursor=None,
                          page_length=1000, neutron_only=True):
    params = []
    if page_cursor:
        params.append("_page_cursor=%s" % page_cursor)
    params.append("_page_length=%s" % page_length)
    # NOTE(salv-orlando): On the NSX backend the 'Quantum' tag is still
    # used for marking Neutron entities in order to preserve compatibility
    if neutron_only:
        params.append("tag_scope=quantum")
    query_params = "&".join(params)
    path = "%s%s%s" % (path, "&" if (path.find("?") != -1) else "?",
                       query_params)
    body = do_request(HTTP_GET, path, cluster=cluster)
    # Result_count won't be returned if _page_cursor is supplied
    return body['results'], body.get('page_cursor'), body.get('result_count')


def get_all_query_pages(path, cluster):
    need_more_results = True
    result_list = []
    page_cursor = None
    while need_more_results:
        results, page_cursor = get_single_query_page(
            path, cluster, page_cursor)[:2]
        if not page_cursor:
            need_more_results = False
        result_list.extend(results)
    return result_list


def mk_body(**kwargs):
    """Convenience function creates and dumps dictionary to string.

    :param kwargs: the key/value pirs to be dumped into a json string.
    :returns: a json string.
    """
    return jsonutils.dumps(kwargs, ensure_ascii=False)



# Copyright 2013 VMware Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import multiprovidernet as mpnet_apidef
from neutron_lib.api.definitions import provider_net as pnet
from neutron_lib.api import validators
from neutron_lib import constants
from neutron_lib import exceptions as n_exc
from oslo_log import log
import six

from vmware_nsx.api_client import client
from vmware_nsx.api_client import exception as api_exc
from vmware_nsx.common import utils as vmw_utils
from vmware_nsx.db import db as nsx_db
from vmware_nsx.db import networkgw_db
from vmware_nsx import nsx_cluster
from vmware_nsx.nsxlib.mh import l2gateway as l2gwlib
from vmware_nsx.nsxlib.mh import router as routerlib
from vmware_nsx.nsxlib.mh import secgroup as secgrouplib
from vmware_nsx.nsxlib.mh import switch as switchlib

LOG = log.getLogger(__name__)


def fetch_nsx_switches(session, cluster, neutron_net_id):
    """Retrieve logical switches for a neutron network.

    This function is optimized for fetching all the lswitches always
    with a single NSX query.
    If there is more than 1 logical switch (chained switches use case)
    NSX lswitches are queried by 'quantum_net_id' tag. Otherwise the NSX
    lswitch is directly retrieved by id (more efficient).
    """
    nsx_switch_ids = get_nsx_switch_ids(session, cluster, neutron_net_id)
    if len(nsx_switch_ids) > 1:
        lswitches = switchlib.get_lswitches(cluster, neutron_net_id)
    else:
        lswitches = [switchlib.get_lswitch_by_id(
            cluster, nsx_switch_ids[0])]
    return lswitches


def get_nsx_switch_ids(session, cluster, neutron_network_id):
    """Return the NSX switch id for a given neutron network.

    First lookup for mappings in Neutron database. If no mapping is
    found, query the NSX backend and add the mappings.
    """
    nsx_switch_ids = nsx_db.get_nsx_switch_ids(
        session, neutron_network_id)
    if not nsx_switch_ids:
        # Find logical switches from backend.
        # This is a rather expensive query, but it won't be executed
        # more than once for each network in Neutron's lifetime
        nsx_switches = switchlib.get_lswitches(cluster, neutron_network_id)
        if not nsx_switches:
            LOG.warning("Unable to find NSX switches for Neutron network "
                        "%s", neutron_network_id)
            return
        nsx_switch_ids = []
        with session.begin(subtransactions=True):
            for nsx_switch in nsx_switches:
                nsx_switch_id = nsx_switch['uuid']
                nsx_switch_ids.append(nsx_switch_id)
                # Create DB mapping
                nsx_db.add_neutron_nsx_network_mapping(
                    session,
                    neutron_network_id,
                    nsx_switch_id)
    return nsx_switch_ids


def get_nsx_switch_and_port_id(session, cluster, neutron_port_id):
    """Return the NSX switch and port uuids for a given neutron port.

    First, look up the Neutron database. If not found, execute
    a query on NSX platform as the mapping might be missing because
    the port was created before upgrading to grizzly.

    This routine also retrieves the identifier of the logical switch in
    the backend where the port is plugged. Prior to Icehouse this
    information was not available in the Neutron Database. For dealing
    with pre-existing records, this routine will query the backend
    for retrieving the correct switch identifier.

    As of Icehouse release it is not indeed anymore possible to assume
    the backend logical switch identifier is equal to the neutron
    network identifier.
    """
    nsx_switch_id, nsx_port_id = nsx_db.get_nsx_switch_and_port_id(
        session, neutron_port_id)
    if not nsx_switch_id:
        # Find logical switch for port from backend
        # This is a rather expensive query, but it won't be executed
        # more than once for each port in Neutron's lifetime
        nsx_ports = switchlib.query_lswitch_lports(
            cluster, '*', relations='LogicalSwitchConfig',
            filters={'tag': neutron_port_id,
                     'tag_scope': 'q_port_id'})
        # Only one result expected
        # NOTE(salv-orlando): Not handling the case where more than one
        # port is found with the same neutron port tag
        if not nsx_ports:
            LOG.warning("Unable to find NSX port for Neutron port %s",
                        neutron_port_id)
            # This method is supposed to return a tuple
            return None, None
        nsx_port = nsx_ports[0]
        nsx_switch_id = (nsx_port['_relations']
                         ['LogicalSwitchConfig']['uuid'])
        if nsx_port_id:
            # Mapping already exists. Delete before recreating
            nsx_db.delete_neutron_nsx_port_mapping(
                session, neutron_port_id)
        else:
            nsx_port_id = nsx_port['uuid']
        # (re)Create DB mapping
        nsx_db.add_neutron_nsx_port_mapping(
            session, neutron_port_id,
            nsx_switch_id, nsx_port_id)
    return nsx_switch_id, nsx_port_id


def get_nsx_security_group_id(session, cluster, neutron_id):
    """Return the NSX sec profile uuid for a given neutron sec group.

    First, look up the Neutron database. If not found, execute
    a query on NSX platform as the mapping might be missing.
    NOTE: Security groups are called 'security profiles' on the NSX backend.
    """
    nsx_id = nsx_db.get_nsx_security_group_id(session, neutron_id)
    if not nsx_id:
        # Find security profile on backend.
        # This is a rather expensive query, but it won't be executed
        # more than once for each security group in Neutron's lifetime
        nsx_sec_profiles = secgrouplib.query_security_profiles(
            cluster, '*',
            filters={'tag': neutron_id,
                     'tag_scope': 'q_sec_group_id'})
        # Only one result expected
        # NOTE(salv-orlando): Not handling the case where more than one
        # security profile is found with the same neutron port tag
        if not nsx_sec_profiles:
            LOG.warning("Unable to find NSX security profile for Neutron "
                        "security group %s", neutron_id)
            return
        elif len(nsx_sec_profiles) > 1:
            LOG.warning("Multiple NSX security profiles found for Neutron "
                        "security group %s", neutron_id)
        nsx_sec_profile = nsx_sec_profiles[0]
        nsx_id = nsx_sec_profile['uuid']
        with session.begin(subtransactions=True):
            # Create DB mapping
            nsx_db.add_neutron_nsx_security_group_mapping(
                session, neutron_id, nsx_id)
    return nsx_id


def get_nsx_router_id(session, cluster, neutron_router_id):
    """Return the NSX router uuid for a given neutron router.

    First, look up the Neutron database. If not found, execute
    a query on NSX platform as the mapping might be missing.
    """
    if not neutron_router_id:
        return
    nsx_router_id = nsx_db.get_nsx_router_id(
        session, neutron_router_id)
    if not nsx_router_id:
        # Find logical router from backend.
        # This is a rather expensive query, but it won't be executed
        # more than once for each router in Neutron's lifetime
        nsx_routers = routerlib.query_lrouters(
            cluster, '*',
            filters={'tag': neutron_router_id,
                     'tag_scope': 'q_router_id'})
        # Only one result expected
        # NOTE(salv-orlando): Not handling the case where more than one
        # port is found with the same neutron port tag
        if not nsx_routers:
            LOG.warning("Unable to find NSX router for Neutron router %s",
                        neutron_router_id)
            return
        nsx_router = nsx_routers[0]
        nsx_router_id = nsx_router['uuid']
        with session.begin(subtransactions=True):
            # Create DB mapping
            nsx_db.add_neutron_nsx_router_mapping(
                session,
                neutron_router_id,
                nsx_router_id)
    return nsx_router_id


def create_nsx_cluster(cluster_opts, concurrent_connections, gen_timeout):
    cluster = nsx_cluster.NSXCluster(**cluster_opts)

    def _ctrl_split(x, y):
        return (x, int(y), True)

    api_providers = [_ctrl_split(*ctrl.split(':'))
                     for ctrl in cluster.nsx_controllers]
    cluster.api_client = client.NsxApiClient(
        api_providers, cluster.nsx_user, cluster.nsx_password,
        http_timeout=cluster.http_timeout,
        retries=cluster.retries,
        redirects=cluster.redirects,
        concurrent_connections=concurrent_connections,
        gen_timeout=gen_timeout)
    return cluster


def get_nsx_device_status(cluster, nsx_uuid):
    try:
        status_up = l2gwlib.get_gateway_device_status(
            cluster, nsx_uuid)
        if status_up:
            return networkgw_db.STATUS_ACTIVE
        else:
            return networkgw_db.STATUS_DOWN
    except api_exc.NsxApiException:
        return networkgw_db.STATUS_UNKNOWN
    except n_exc.NotFound:
        return networkgw_db.ERROR


def get_nsx_device_statuses(cluster, tenant_id):
    try:
        status_dict = l2gwlib.get_gateway_devices_status(
            cluster, tenant_id)
        return dict((nsx_device_id,
                     networkgw_db.STATUS_ACTIVE if connected
                     else networkgw_db.STATUS_DOWN) for
                    (nsx_device_id, connected) in six.iteritems(status_dict))
    except api_exc.NsxApiException:
        # Do not make a NSX API exception fatal
        if tenant_id:
            LOG.warning("Unable to retrieve operational status for "
                        "gateway devices belonging to tenant: %s",
                        tenant_id)
        else:
            LOG.warning("Unable to retrieve operational status for "
                        "gateway devices")


def _convert_bindings_to_nsx_transport_zones(bindings):
    nsx_transport_zones_config = []
    for binding in bindings:
        transport_entry = {}
        if binding.binding_type in [vmw_utils.NetworkTypes.FLAT,
                                    vmw_utils.NetworkTypes.VLAN]:
            transport_entry['transport_type'] = (
                vmw_utils.NetworkTypes.BRIDGE)
            transport_entry['binding_config'] = {}
            vlan_id = binding.vlan_id
            if vlan_id:
                transport_entry['binding_config'] = (
                    {'vlan_translation': [{'transport': vlan_id}]})
        else:
            transport_entry['transport_type'] = binding.binding_type
        transport_entry['zone_uuid'] = binding.phy_uuid
        nsx_transport_zones_config.append(transport_entry)
    return nsx_transport_zones_config


def _convert_segments_to_nsx_transport_zones(segments, default_tz_uuid):
    nsx_transport_zones_config = []
    for transport_zone in segments:
        for value in [pnet.NETWORK_TYPE, pnet.PHYSICAL_NETWORK,
                      pnet.SEGMENTATION_ID]:
            if transport_zone.get(value) == constants.ATTR_NOT_SPECIFIED:
                transport_zone[value] = None

        transport_entry = {}
        transport_type = transport_zone.get(pnet.NETWORK_TYPE)
        if transport_type in [vmw_utils.NetworkTypes.FLAT,
                              vmw_utils.NetworkTypes.VLAN]:
            transport_entry['transport_type'] = (
                vmw_utils.NetworkTypes.BRIDGE)
            transport_entry['binding_config'] = {}
            vlan_id = transport_zone.get(pnet.SEGMENTATION_ID)
            if vlan_id:
                transport_entry['binding_config'] = (
                    {'vlan_translation': [{'transport': vlan_id}]})
        else:
            transport_entry['transport_type'] = transport_type
        transport_entry['zone_uuid'] = (
            transport_zone[pnet.PHYSICAL_NETWORK] or default_tz_uuid)
        nsx_transport_zones_config.append(transport_entry)
    return nsx_transport_zones_config


def convert_to_nsx_transport_zones(
    default_tz_uuid, network=None, bindings=None,
    default_transport_type=None):

    # Convert fields from provider request to nsx format
    if (network and not validators.is_attr_set(
        network.get(mpnet_apidef.SEGMENTS))):
        return [{"zone_uuid": default_tz_uuid,
                 "transport_type": default_transport_type}]

    # Convert fields from db to nsx format
    if bindings:
        return _convert_bindings_to_nsx_transport_zones(bindings)

    # If we end up here we need to convert multiprovider segments into nsx
    # transport zone configurations
    return _convert_segments_to_nsx_transport_zones(
        network.get(mpnet_apidef.SEGMENTS), default_tz_uuid)



# Copyright 2017 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Update the primary key constraint of nsx_subnet_ipam

Revision ID: 8c0a81a07691
Revises: 14a89ddf96e2
Create Date: 2017-02-15 15:25:21.163418

"""

# revision identifiers, used by Alembic.
revision = '8c0a81a07691'
down_revision = '14a89ddf96e2'

from alembic import op
from sqlalchemy.engine import reflection


def upgrade():
    table_name = 'nsx_subnet_ipam'
    # replace the old primary key constraint with a new one for both
    # subnet and nsx-pool
    inspector = reflection.Inspector.from_engine(op.get_bind())
    pk_constraint = inspector.get_pk_constraint(table_name)
    op.drop_constraint(pk_constraint.get('name'), table_name, type_='primary')
    op.create_primary_key(None, table_name,
                          ['subnet_id', 'nsx_pool_id'])
